###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from copy import copy, deepcopy
from inspect import isroutine
from types import GenericAlias, UnionType
from typing import Any, Iterator, Self
from _collections_abc import dict_items, dict_keys, dict_values

from everysk.core.datetime import DateTime, Date
from everysk.core.exceptions import DefaultError, RequiredError, FieldValueError


CLASS_KEY: str = '__class_path__'
CONFIG_ATTRIBUTE_NAME: str = '_config'


###############################################################################
#   Private functions Implementation
###############################################################################
def __get_field_value__(obj: Any, attr: str, value: Any) -> Any:
    """
    Function that get the cleaned value for a Field and validate this value.

    Args:
        obj (Any): A class or a instance of BaseObject.
        attr (str): The attribute name.
        value (Any): The value that is assigned to this attribute.

    Raises:
        FieldValueError: If we find validation errors.
    """
    # Transforms the given value to Undefined if it matches the default_parse_string.
    value = _transform_to_undefined(value)

    # Get all attributes that the object has
    attributes = getattr(obj, MetaClass._attr_name) # pylint: disable=protected-access
    try:
        field = attributes[attr]
    except KeyError:
        return value

    # field can be the type itself or an instance of BaseField
    if isinstance(field, BaseField):
        try:
            value = field.get_cleaned_value(value)
        except Exception as error:
            # Add attribute name to error
            error.args = (f'{attr}: {str(error)}',)
            raise FieldValueError(error.args) from error

        field.validate(attr, value)
    else:
        _validate(attr, value, field)

    return value

def _transform_to_undefined(value: Any) -> Any:
    """
    Transforms the given value to Undefined if it matches the default parse string.

    Args:
        value (Any): The value to transform.

    Returns:
        Any: The transformed value.
    """
    if isinstance(value, str) and value == Undefined.default_parse_string:
        value = Undefined

    return value

def _required(attr_name: str, value: Any) -> None:
    """
    Checks if value is required, required values can't be: None, '', [], {}.

    Raises:
        RequiredError: When value is required and match with False conditions.
    """
    if value in (Undefined, None, '', (), [], {}):
        raise RequiredError(f'The {attr_name} attribute is required.')

def _validate(attr_name: str, value: Any, attr_type: type | UnionType) -> None: # pylint: disable=too-many-branches, too-many-return-statements
    """
    Validates that the given value is of the expected attribute type. The function supports special type checks for
    Date and DateTime types and handles general type validation for other types. It allows the value to pass through
    if it matches the expected type, is None, or is an instance of Undefined. The function uses custom checks for Date
    and DateTime to accommodate their unique validation requirements.

    Args:
        attr_name (str):
            The name of the attribute being validated. This is used for generating error messages.

        value (Any):
            The value to be validated against the expected type.

        attr_type (type):
            The expected type of the value. This can be a standard Python type, a custom type, or
            specific types like Date and DateTime which have dedicated validation logic.

    Raises:
        FieldValueError: When value is not of value_type.
    """
    # We always accept these 2 values
    if value is None or value is Undefined:
        return None

    if attr_type == Date:
        # If we are expecting a Date we don't need to check other things
        if Date.is_date(value):
            return None

    if attr_type == DateTime:
        # If we are expecting a DateTime we don't need to check other things
        if DateTime.is_datetime(value):
            return None

    # TypeError: typing.Any cannot be used with isinstance()
    if attr_type is Any:
        return None

    # https://everysk.atlassian.net/browse/COD-4286
    # If we use 'class'/callable as a annotation, the isinstance will fail
    # because attr_type will be a string/function so we need to discard it first
    if attr_type is callable:
        if callable(value):
            return None

    # If it is string, when we use classes as annotations we check if the name of the class is the same
    if isinstance(attr_type, str):
        if type(value).__name__ == attr_type:
            return None

    try:
        # Check if value is an instance of the expected type
        # If attr_type is a UnionType - int | float, the isinstance will work
        if isinstance(value, attr_type):
            return None
    except TypeError:
        pass

    ## Subscriptable types like List, Dict, Tuple, etc
    if isinstance(attr_type, GenericAlias):
        # We need to check if the value is a instance of the origin type
        if isinstance(value, attr_type.__origin__):
            # We not check the content of the list, dict, tuple, etc
            return None

    raise FieldValueError(f'Key {attr_name} must be {attr_type}.')


###############################################################################
#   BaseField Class Implementation
###############################################################################
class BaseField:
    """ Base class of all fields that will guarantee their type. """
    ## Public attributes
    attr_type: type | UnionType = None
    default: Any = None
    readonly: bool = False
    required: bool = False
    required_lazy: bool = False
    empty_is_none: bool = False

    def __init__(
        self,
        attr_type: type | UnionType = None,
        default: Any = None,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        """
        Use kwargs to set more attributes on the Field.

        Raises:
            DefaultError: For default values they can't be empty [] or {}.
            RequiredError: If field is readonly, then default value is required.
        """
        self.attr_type = attr_type
        if required and required_lazy:
            raise FieldValueError("Required and required_lazy can't be booth True.")

        self.required = required
        self.required_lazy = required_lazy

        if default is not None and not default and isinstance(default, (list, dict)):
            # For default values they can't be empty [] or {} - because this can cause
            # some issues with class attributes where these type can aggregate values.
            raise DefaultError('Default value cannot be a list or a dict.')

        if readonly and (default is None or default is Undefined):
            raise RequiredError('If field is readonly, then default value is required.')

        self.readonly = readonly

        # We use this flag to convert '' to None
        self.empty_is_none = empty_is_none

        # Other attributes will be assigned directly
        for key, value in kwargs.items():
            setattr(self, key, value)

        # For the last We need to store the cleaned value
        self.default = self.get_cleaned_value(default)

    def __repr__(self) -> str:
        """
        The `__repr__` method from `BaseField` returns the name of the
        instantiated class.

        Returns:
            str: The name of the class.
        """
        return self.__class__.__name__

    def __eq__(self, obj: object) -> bool:
        """
        One object will be equal to another one if all `__dict__`
        attributes are the same.

        Args:
            obj (object): The obj for comparison.
        """
        return self.__dict__ == obj.__dict__

    def transform_to_none(self, value: Any) -> Any:
        """
        Transforms value to None if needed.

        Args:
            value (Any): The value to be converted to `None`.

        Example:
            >>> field = BaseField(attr_type=str, empty_is_none=False)
            >>> field.transform_to_none('')
            ''

            >>> field = BaseField(attr_type=str, empty_is_none=True)
            >>> field.transform_to_none('')
            None
        """
        if self.empty_is_none and value == '':
            value = None

        return value

    def get_cleaned_value(self, value: Any) -> Any:
        """
        This function first converts the value to None if needed, then
        checks if the `value` is a callable. If it is, the function calls
        the `value` and stores its result by reassigning the `value` variable.
        Finally, the function calls the `clean_value` method.

        Args:
            value (Any): The cleaned value to be retrieved.

        Returns:
            Any: The cleaned value.
        """
        # We first verify if we need to transform the value to None
        value = self.transform_to_none(value)

        # Then we run the get_value method when value is a callable
        value = self.get_value(value)

        # Then we run the clean_value method
        value = self.clean_value(value)

        return value

    def get_value(self, value: Any) -> Any:
        """
        This function checks if the `value` is a callable
        By either returning the `value` or returning the
        result of the callable function.

        Must be implemented in child classes that need do some changes
        on received value before clean_value.

        Args:
            value (Any): The value to be possibly called.

        Returns:
            Any: The result of the callable function or the original value.
        """
        # If value is callable we call it
        if callable(value):
            value = value()

        return value

    def clean_value(self, value: Any) -> Any:
        """
        This method is always called when we assigned a value to some attribute.
        Must be implemented in child classes to change the behavior of a given field.
        Below we have an example that reimplements the `clean_value()`.

        Args:
            value (Any): The value to be cleaned.

        Usage:
            >>> from everysk.core.fields import BoolField
            >>> from everysk.core.object import BaseObject

            >>> class MyBoolField(BoolField):
            ...     def clean_value(self, value):
            ...         if value == 'test':
            ...             return True
            ...         return False

            >>> class MyClass(BaseObject):
            ...     f1 = MyBoolField()

            >>> a = MyClass()
            >>> a.f1 = True
            >>> a.f1
            False

            >>> a.f1 = 'test'
            >>> a.f1
            True
        """
        return value

    def validate(self, attr_name: str, value: Any, attr_type: type | UnionType = None) -> None:
        """
        Checks if value is required and if is of correct type.
        This method can be reimplemented in child classes to modify
        the behavior.

        Raises:
            RequiredError: If required and None is passed
            FieldValueError: If value type don't match with required type.

        Usage:
            >>> from everysk.core.fields import StrField
            >>> from everysk.core.object import BaseObject

            >>> class MyStrField(StrField):
            ...     def validate(self, attr_name, value):
            ...         if value == 'test':
            ...             raise ValueError("value argument cannot be 'test'")

            >>> class MyClass(BaseObject):
            ...     f1 = MyStrField()

            >>> a = MyClass()
            >>> a.f1 = 'test'
            ValueError: value argument cannot be 'test'
        """
        if attr_type is None:
            attr_type = self.attr_type

        if self.readonly:
            # This is necessary to be able to at least assign the default value to the field
            if value != self.default:
                raise FieldValueError(f"The field '{attr_name}' value cannot be changed.")

        if self.required and not self.required_lazy:
            _required(attr_name, value)

        _validate(attr_name, value, attr_type)

    def __getattr__(self, name: str) -> Any:
        """
        This method is used to handle pylint errors where the method/attribute does not exist.
        The problem is that we change the field in the MetaClass to the Field's default value,
        so StrField does not have the str methods but the result is a string.
        This method will only be executed if the method/attributes do not exist in the Field class.

        Args:
            name (str): The name of the method/attribute.
        """
        # https://pythonhint.com/post/2118347356810295/avoid-pylint-warning-e1101-instance-of-has-no-member-for-class-with-dynamic-attributes
        if not isinstance(self.attr_type, UnionType):
            return getattr(self.attr_type, name)

        # We try to get the attribute from the ones that are in the tuple
        for attr_type in self.attr_type.__args__:
            try:
                return getattr(attr_type, name)
            except AttributeError:
                pass

        # If no attribute was found we raise the error
        raise AttributeError(f"type object '{self.attr_type}' has no attribute '{name}'.")


###############################################################################
#   MetaClass Implementation
###############################################################################
def _silent(func: callable) -> callable:
    """
    Function that creates a silent decorator for the given function.
    This decorator will catch any exception raised by the function and store it.

    Args:
        func (callable): The function to be decorated.
    """
    def wrapper(self, *args, **kwargs):
        # pylint: disable=broad-exception-caught, protected-access
        silent = kwargs.pop('silent', self._silent)
        try:
            return func(self, *args, **kwargs)
        except Exception as error:
            if not silent:
                raise error
            if self._errors is None:
                self._errors = {}
            self._errors['init'] = deepcopy(error)

        return None

    return wrapper

class MetaClass(type):
    _attr_name: str = '__attributes__'
    _anno_name: str = '__annotations__'

    def __call__(cls, *args: tuple, **kwargs: dict) -> Self:
        """
        Method that creates a sequence of class call like:
            before_init
            init
            after_init
        If the BaseObject class implement one of these methods they will be executed.
        https://discuss.python.org/t/add-a-post-method-equivalent-to-the-new-method-but-called-after-init/5449/11
        """
        errors: dict[str: Exception | None] = {'before_init': None, 'init': None, 'after_init': None}
        # We get the silent from kwargs otherwise we get from the class
        silent = kwargs.get('silent', cls._silent)

        ## If the before init method is implemented we run it
        try:
            # If the method returns a dict we update kwargs
            dct = cls.__before_init__(**kwargs)
            if isinstance(dct, dict):
                kwargs = dct
        except Exception as error: # pylint: disable=broad-exception-caught
            if not silent:
                raise error
            errors['before_init'] = deepcopy(error)

        ## Here we create the object and initialize the silent for init must be inside the _silent function
        obj = super().__call__(*args, **kwargs)
        if obj._errors:
            errors['init'] = obj._errors['init']

        ## If the after init method is implemented we run it
        try:
            obj.__after_init__()
        except Exception as error: # pylint: disable=broad-exception-caught
            if not silent:
                raise error
            errors['after_init'] = deepcopy(error)

        # Store the errors
        if any(errors.values()):
            obj._errors = errors
            # Execute the handler for errors
            obj._init_error_handler(kwargs, errors)

        # Set the instance to be Frozen or not
        try:
            config = getattr(obj, CONFIG_ATTRIBUTE_NAME)
            obj._is_frozen = config.frozen # pylint: disable=attribute-defined-outside-init
        except AttributeError:
            pass

        return obj

    def __new__(mcs, name: str, bases: tuple, attrs: dict) -> Self:
        """
        This method is executed every time a BaseObject Class is created in the Python runtime.
        We changed this method to create the config and attributes properties and update the annotations.

        Example:
            >>> from everysk.core.object import BaseObject, CONFIG_ATTRIBUTE_NAME
            >>> class MyClass(BaseObject):
            ...     class Config:
            ...         value: int = 1

            >>> obj1 = MyClass()
            >>> obj2 = MyClass()
            >>> configC = getattr(MyClass, CONFIG_ATTRIBUTE_NAME)
            >>> config1 = getattr(obj1, CONFIG_ATTRIBUTE_NAME)
            >>> config2 = getattr(obj2, CONFIG_ATTRIBUTE_NAME)
            >>> configC == config1 == config2
            ... True

            >>> configC.value, config1.value, config2.value
            (1, 1, 1)

            >>> config2.value = 3
            >>> configC.value, config1.value, config2.value
            (3, 3, 3)

            >>> MyClass.Config
            ---------------------------------------------------------------------------
            AttributeError                            Traceback (most recent call last)
            ----> 1 MyClass.Config

            AttributeError: type object 'MyClass' has no attribute 'Config'

        Args:
            mcs (type): Represents this class.
            name (str): The name for the new class.
            bases (tuple): All inheritance classes.
            attrs (dict): All attributes that the new class has.

        Returns:
            type: Return the new class object.
        """
        # Creating the config property.
        if 'Config' in attrs:
            # If Config is in the class we just create it
            Config = attrs.pop('Config') # pylint: disable=invalid-name
            attrs[CONFIG_ATTRIBUTE_NAME] = Config()
        else:
            # If Config is not in the class we get the first one from the bases
            for base in bases:
                if hasattr(base, CONFIG_ATTRIBUTE_NAME):
                    config = getattr(base, CONFIG_ATTRIBUTE_NAME)
                    attrs[CONFIG_ATTRIBUTE_NAME] = deepcopy(config)
                    break

        # We need all attributes to validate the types later
        # So we need to get the parents attributes too
        attributes = {}
        for parent in bases:
            attributes.update(getattr(parent, mcs._attr_name, {}))

        # We could not update the info inside the original attrs dict because:
        # RuntimeError: dictionary changed size during iteration
        # So we remove the attributes that we need to update
        attributes.update(attrs.pop(mcs._attr_name, {}))
        annotations: dict = attrs.pop(mcs._anno_name, {})
        new_attrs = {}
        for attr_name, attr_value in attrs.items():
            if attr_name == '__init__':
                # To run all init in silent mode we need to decorate it
                new_attrs[attr_name] = _silent(attr_value)

            # We discard all python dunder attributes, all functions, all properties, Undefined and None values
            elif not attr_name.startswith('__') and \
               not isroutine(attr_value) and \
               not isinstance(attr_value, property) and \
               attr_value is not None and \
               attr_value is not Undefined:
                # For BaseFields we need to use the properties and set the correct value in the attribute
                if isinstance(attr_value, BaseField):
                    # We keep a copy of the value inside the __attributes__
                    attributes[attr_name] = attr_value

                    # We set the correct value to this attribute in the class
                    new_attrs[attr_name] = attr_value.default

                    # We create the annotation for this attribute
                    if attr_name not in annotations:
                        annotations[attr_name] = attr_value.attr_type
                else:
                    # For normal attributes we only store the class
                    attributes[attr_name] = type(attr_value)

                    # Now we update annotations for attributes that are not annotated
                    # Ex: var = 1
                    if attr_name not in annotations:
                        annotations[attr_name] = type(attr_value)

        # With both completed now we need to get the fields that are only annotations
        # class MyClass:
        #     var: str
        for key in annotations.keys() - attributes.keys():
            attributes[key] = annotations[key]
            # We set the default value to None to avoid break the code
            new_attrs[key] = None

        # We remove config from these to avoid validation/serialization problems
        if CONFIG_ATTRIBUTE_NAME in attributes:
            del attributes[CONFIG_ATTRIBUTE_NAME]

        if CONFIG_ATTRIBUTE_NAME in annotations:
            del annotations[CONFIG_ATTRIBUTE_NAME]

        # Readonly attributes need to go in the exclude list to generate the to_dict result correctly
        if CONFIG_ATTRIBUTE_NAME in attrs and hasattr(attrs[CONFIG_ATTRIBUTE_NAME], 'exclude_keys'):
            readonly_keys = {key for key, value in attributes.items() if getattr(value, 'readonly', False)}
            attrs[CONFIG_ATTRIBUTE_NAME].exclude_keys = attrs[CONFIG_ATTRIBUTE_NAME].exclude_keys.union(readonly_keys)

        # Then we update the attributes list for this new class
        attrs[mcs._attr_name] = attributes
        attrs[mcs._anno_name] = annotations
        attrs.update(new_attrs)

        return super().__new__(mcs, name, bases, attrs)

    def __setattr__(cls, __name: str, __value: Any) -> None:
        """
        Method that sets the values on fields of the class.

        Args:
            __name (str): The attribute name.
            __value (Any): The value that is set.
        """
        return super().__setattr__(__name, __get_field_value__(cls, __name, __value))


###############################################################################
#   BaseObject Class Implementation
###############################################################################
class _BaseObject(metaclass=MetaClass):
    """
    To ensure correct check for data keys
    uses https://docs.python.org/3/library/typing.html standards.

        >>> from utils.object import BaseObject
        >>> class MyObject(BaseObject):
        ...     var_int: int = None
        ...
        >>> obj = MyObject(var_int='a')
        Traceback (most recent call last):
        File "<stdin>", line 1, in <module>
        File "/var/app/utils/object.py", line 151, in __init__
            setattr(self, key, value)
        File "/var/app/utils/object.py", line 209, in __setattr__
            value = self.__get_clean_value__(__name, value)
        File "/var/app/utils/object.py", line 192, in __get_clean_value__
            _validate(attr, value, annotations[attr])
        File "/var/app/utils/object.py", line 56, in _validate
            raise DataTypeError(f'Key {attr_name} must be {attr_type}.')
        utils.exceptions.DataTypeError: Key var_int must be <class 'int'>.
    """
    ## Private attributes
    __slots__ = ('_need_validation',) # If we need to validate the data on setattr.
    _errors: dict = None # Keep the init errors.
    _is_frozen: bool = False # This will control if we can update data on this class.
    _silent: bool = False # If true an error that happen on init will be stored in self._error.

    def __new__(cls, *args, **kwargs) -> Self:
        obj = super().__new__(cls)

        # Initialize the _need_validation attribute always as True to validate all fields
        if not hasattr(obj, '_need_validation') or obj._need_validation is None:
            obj._need_validation = True

        return obj

    @classmethod
    def __before_init__(cls, **kwargs: dict) -> dict:
        """
        Method that runs before the __init__ method.
        If needed we must return the kwargs that will be passed to the init
        otherwise return None to not change the current behavior.
        """
        return kwargs

    def __init__(self, **kwargs: dict) -> None:
        # Validate all required fields
        attributes = self.__get_attributes__()
        for attr_name, field in attributes.items():
            if getattr(field, 'required', False):
                _required(attr_name=attr_name, value=kwargs.get(attr_name))

        # Set all kwargs on the object
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __after_init__(self) -> None:
        """
        Method that runs after the __init__ method.
        This method must return None.
        """
        pass

    def __check_frozen__(self) -> None:
        """
        Method that checks if this class is a Frozen object and raises attribute error.
        """
        if self._is_frozen:
            raise AttributeError(f'Class {self.get_full_doted_class_path()} is frozen and cannot be modified.')

    def __copy__(self) -> Self:
        """
        A shallow copy constructs a new compound object and then (to the extent possible)
        inserts references into it to the objects found in the original.
        If the object is Frozen the copy will not be.
        This method is used when we call copy(obj).
        """
        # We need to copy the __dict__
        obj = copy(self.__dict__)

        # We must not copy the config attribute because the copy could override the original in the class
        obj.pop(CONFIG_ATTRIBUTE_NAME, None)

        # We create a new obj
        obj = type(self)(**obj)
        return obj

    def __deepcopy__(self, memo: dict = None) -> Self:
        """
        A deep copy constructs a new compound object and then, recursively,
        inserts copies into it of the objects found in the original.
        This method is used when we call deepcopy(obj).

        Args:
            memo (dict, optional): A memory object to avoid copy twice. Defaults to None.
        """
        # We need to copy the __dict__
        obj = deepcopy(self.__dict__, memo)

        # We must not copy the config attribute because the copy could override the original in the class
        obj.pop(CONFIG_ATTRIBUTE_NAME, None)

        # We create a new obj
        obj = type(self)(**obj)
        return obj

    def __delattr__(self, __name: str) -> None:
        """
        Method that removes __name from the object.

        Args:
            __name (str): The name of the attribute that will be removed.

        Raises:
            AttributeError: If is frozen.
        """
        self.__check_frozen__()
        super().__delattr__(__name)

    def __get_attributes__(self) -> dict:
        """
        Get all attributes from this class.
        """
        return getattr(self, MetaClass._attr_name) # pylint: disable=protected-access

    def __get_clean_value__(self, attr: str, value: Any) -> Any:
        """
        Pass value to a clean function and checks if value match's the correct type.

        Raises:
            FieldValueError: If value and type don't match the correct type.
        """
        return __get_field_value__(self, attr, value)


    def _init_error_handler(self, kwargs: dict, errors: dict[str, Exception]) -> None:
        """
        This method is called at the end of the init process if the silent param is True.

        Args:
            errors (dict): A dict {'before_init': None | Exception, 'init': None | Exception, 'after_init': None | Exception}
        """
        pass

    def __setattr__(self, __name: str, __value: Any) -> None:
        """
        Method changed from BaseClass for check de integrity of data.
        This method is executed on setting attributes in the object.
        Ex: obj.attr = 1

        Raises:
            AttributeError: If is frozen.
        """
        self.__check_frozen__()

        if getattr(self, '_need_validation', True):
            __value = self.__get_clean_value__(__name, __value)

        super().__setattr__(__name, __value)

    @classmethod
    def __set_attribute__(cls, attr_name: str, attr_type: Any, attr_value: Any) -> None:
        """
        Method that updates the list of attributes/annotations for this class.
        Normally this is used to update a class after it was created.

        Args:
            attr_name (str): The name for the new attribute.
            attr_type (Any): The type for the new attribute.
            attr_value (Any): The value for the new attribute.
        """
        attributes = getattr(cls, MetaClass._attr_name, {}) # pylint: disable=protected-access
        annotations = getattr(cls, MetaClass._anno_name, {}) # pylint: disable=protected-access
        try:
            # For BaseFields
            annotations[attr_name] = attr_type.attr_type
        except AttributeError:
            # Normal types
            annotations[attr_name] = attr_type

        attributes[attr_name] = attr_type
        # After we set the attributes/annotations we set the value in the class
        setattr(cls, attr_name, attr_value)

    ## Public methods
    def get_full_doted_class_path(self) -> str:
        """
        Return full doted class path to be used on import functions.

        Example:
            'everysk.core.BaseObject'
        """
        return f'{self.__module__}.{self.__class__.__name__}'

    def replace(self, **changes) -> Self:
        """
        Creates a new object of the same type as self, replacing fields with values from changes.

        Args:
            **changes (dict): All named params that are passed to this method.

        Returns:
            Self: A copy of this object with the new values.

        Example:
            >>> from everysk.core.object import BaseObject
            >>> obj = BaseObject(attr=1)
            >>> obj.attr
            1
            >>> copy = obj.replace(attr=2)
            >>> copy.attr
            2
        """
        obj = deepcopy(self.__dict__)

        # We must not copy the config attribute because the copy could override the original in the class
        obj.pop(CONFIG_ATTRIBUTE_NAME, None)

        obj.update(changes)
        return type(self)(**obj)

    def validate_required_fields(self) -> None:
        """
        Try to validate all fields that are checked with required_lazy, because fields with
        required are always validate on the init.
        """
        attributes = self.__get_attributes__()
        for attr_name, field in attributes.items():
            if getattr(field, 'required_lazy', False):
                _required(attr_name=attr_name, value=getattr(self, attr_name, None))


###############################################################################
#   BaseObjectConfig Implementation
###############################################################################
class BaseObjectConfig(_BaseObject):
    exclude_keys: frozenset[str] = frozenset([])
    key_mapping: dict = None

    def __init__(self, **kwargs: dict) -> None:
        """ Use kwargs to set more attributes on the Config. """
        super().__init__(**kwargs)
        if self.key_mapping is None:
            self.key_mapping = {}


###############################################################################
#   BaseObject Class Implementation
###############################################################################
class BaseObject(_BaseObject):
    class Config(BaseObjectConfig):
        pass

    ## Private attributes
    _config: Config = None

    def __getstate__(self) -> dict:
        """
        This method is used by Pickle module to get the correct serialized data.
        https://docs.python.org/3.11/library/pickle.html#handling-stateful-objects
        """
        # This generates a dictionary with all attributes
        # We need to get the keys that are set in the instance and in the parents
        keys = self.__dict__.keys() | self.__get_attributes__().keys()

        # Then config key need to be removed and the exclude_keys list too
        config = getattr(self, CONFIG_ATTRIBUTE_NAME)
        keys = keys - {CONFIG_ATTRIBUTE_NAME} - set(config.exclude_keys)

        dct: dict = {key: getattr(self, key) for key in keys}

        # We set the _need_validation attribute to False because we already validated all data
        dct['_need_validation'] = False

        return dct

    def __setstate__(self, state: dict = None) -> None:
        """
        This method is used by Pickle module to set back the correct serialized data.
        We need to iterate over every key and set the value to the object.

        Args:
            state (dict): The result from the __getstate__ method used by Pickle.
        """
        if state:
            old_need_validation = self._need_validation
            # Set if we need validate every attribute
            self._need_validation = state.pop('_need_validation', True)
            for key, value in state.items():
                setattr(self, key, value)
            # Set the original value to validate the attributes
            self._need_validation = old_need_validation

    def to_native(self, add_class_path: str | None = None, recursion: bool = False) -> Any:
        """
        Converts the object to the specified Python type.

        Args:
            add_class_path (str | None, optional): The class path to add when converting the object. Defaults to None.
            recursion (bool, optional): Indicates whether to recursively convert nested objects. Defaults to False.

        Returns:
            object: The converted object.

        """
        return self.to_dict(add_class_path=add_class_path, recursion=recursion)

    def to_dict(self, add_class_path: bool = False, recursion: bool = False) -> dict:
        """
        This method is used to convert the object to a dictionary.
        If add_class_path is True, the full doted class path will be added to the dictionary.
        If recursion is True, the method will call the to_dict method of the child objects.

        Args:
            add_class_path (bool, optional): Flag to add the class path in the result. Defaults to False.
            recursion (bool, optional): Flag to transform the children too. Defaults to False.
        """
        dct: dict = {}
        # We need to get the keys that are set in the instance
        keys = self.__dict__.keys()
        if add_class_path:
            # If add_class_path is True we need to add all other keys that are in the class with the default value
            keys = keys | self.__get_attributes__().keys()

        def no_op(value: Any) -> Any:
            """ Function used to return the value in the getattr. """
            return value

        # Then config key need to be removed and the exclude_keys too
        config = getattr(self, CONFIG_ATTRIBUTE_NAME)
        keys = keys - {CONFIG_ATTRIBUTE_NAME} - set(config.exclude_keys)

        for key in keys:
            value = getattr(self, key)
            func = getattr(self, f'_process_{key}', no_op)
            key = config.key_mapping.get(key, key)
            result = func(value)
            if recursion and isinstance(result, BaseObject):
                result = result.to_dict(add_class_path=add_class_path, recursion=recursion)

            dct[key] = result

        if add_class_path:
            dct[CLASS_KEY] = self.get_full_doted_class_path()

        return dct


###############################################################################
#   BaseDictConfig Implementation
###############################################################################
class BaseDictConfig(BaseObjectConfig):
    keys_blacklist: frozenset[str] = frozenset([])


###############################################################################
#   BaseDict Class Implementation
###############################################################################
class BaseDict(BaseObject):
    """
    Extends BaseObject and also guarantees that BaseDict['key'] is equal to BaseDict.key

        >>> from utils.object import BaseDict
        >>> class MyDict(BaseDict):
        ...     var_int: int = None
        ...
        >>> obj = MyDict(var_int='test')
        Traceback (most recent call last):
        File "<stdin>", line 1, in <module>
        File "/var/app/utils/object.py", line 151, in __init__
            raise DataTypeError(f'Key {attr_name} must be {attr_type}.')
        File "/var/app/utils/object.py", line 279, in __setattr__
            if key.startswith('_'):
        File "/var/app/utils/object.py", line 209, in __setattr__
            DataTypeError: If value and type don't match the correct type.
        File "/var/app/utils/object.py", line 192, in __get_clean_value__
            # Create base annotations
        File "/var/app/utils/object.py", line 56, in _validate
            elif isinstance(value, attr_type):
        utils.exceptions.DataTypeError: Key var_int must be <class 'int'>.
        >>> obj = MyDict(var_int=10)
        >>> obj['var_int'] == obj.var_int
        True
    """
    class Config(BaseDictConfig):
        pass

    ## Private attributes
    __slots__ = ('__data__',)
    _config: Config = None

    ## Private Methods
    def __new__(cls, *args, **kwargs) -> Self:
        obj = super().__new__(cls, *args, **kwargs)
        # We need to create this here because __init__ could crash if we use self.attr
        # before call the super or if we use silent=True and work with the object later
        # Create the __data__ attribute if it does not exist
        if not hasattr(obj, '__data__'):
            obj.__data__ = {}

        return obj

    def __init__(self, **kwargs) -> None:
        # This is only for pylint stop complaining about not having __dict__
        if not hasattr(self, '__dict__'):
            self.__dict__ = {}

        # Add all kwargs to the object as key/attributes
        super().__init__(**kwargs)

        # We need to get all attributes from the parent classes too
        # to work like a dict and to have correct equality checks and
        # to represent the object as a dict
        attributes = self.__get_attributes__()
        for key in attributes.keys() - kwargs.keys():
            if self.is_valid_key(key=key):
                self[key] = self[key]

    def __contains__(self, key: str) -> bool:
        """
        Check if key is in self or in a parent.

        Args:
            key (str): The key name to search.
        """
        return key in self.__data__

    def __delattr__(self, name: str, caller: str = None) -> None:
        """
        Removes an atribute from self.

        Args:
            name (str): The attribute that will be removed.
            caller (str, optional): Used to avoid recursion when internally called. Defaults to None.

        Raises:
            AttributeError: If the attribute is not found.
        """
        # Remove attribute
        super().__delattr__(name)

        # Caller not None means this method was called from __delitem__
        # then we do not call the method again avoiding infinite loop
        if caller is None:
            try:
                # Some times the attribute will not exists
                self.__delitem__(name, caller='__delattr__')
            except KeyError:
                pass

    def __delitem__(self, key: Any, caller: str = None) -> None:
        """
        Removes an key from self.

        Args:
            key (Any): The key that will be removed.
            caller (str, optional): Used to avoid recursion when internally called. Defaults to None.

        Raises:
            KeyError: If the key is not found.
        """
        # Remove key
        self.__data__.__delitem__(key)

        # Caller not None means this method was called from __delattr__
        # then we do not call the method again avoiding infinite loop
        if caller is None:
            self.__delattr__(key, caller='__delitem__')

    def __eq__(self, other: Any) -> bool:
        """
        Check if two objects are the same.

        Args:
            other (Any): The other object to compare.
        """
        return isinstance(other, BaseDict) and self.__data__ == other.__data__

    def __getattr__(self, name: str) -> Any:
        """
        If the self does not have the attribute this method is called.
        We check if the __data__ does not have the attribute.

        Args:
            name (str): The name of the desired attribute.

        Raises:
            AttributeError: If name is not found.
        """
        if name != '__data__':
            try:
                return getattr(self.__data__, name)
            except AttributeError:
                pass

        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'.")

    def __getitem__(self, key: str) -> Any:
        """
        Get key from self or search on a parent.

        Args:
            key (str): The key name to search.

        Raises:
            KeyError: If the key does not exist.
        """
        if self.is_valid_key(key=key):
            try:
                return self.__data__.__getitem__(key)
            except KeyError:
                pass

            try:
                return getattr(self, key)
            except AttributeError:
                pass

        raise KeyError(key) # pylint: disable=raise-missing-from

    def __ior__(self, other: Any) -> Any:
        """
        This method performs an in-place merge of two objects using the `|=` operator.
        If `other` is an instance of the same class, it merges the attributes of `other` into the `self` class.
        Otherwise, it tries to merge the `other` dictionary into the `self` class.

        Args:
            other (Any): The other object to merge.

        Returns:
            Any: The instance of the class with the merged attributes.
        """
        if isinstance(other, type(self)):
            self.__dict__ |= other.__dict__

        else:
            self.__dict__ |= other

        return self

    def __iter__(self) -> Iterator:
        """
        Iterate over the object keys.
        This method is used when we call a for over the instance:

        Example:
            >>> for key in BaseDict():
        """
        return iter(self.__data__)

    def __len__(self) -> int:
        """
        Get the length of the dictionary.

        Returns:
            int: The length of the dictionary.
        """
        return len(self.__data__)

    def __ne__(self, other: Any) -> bool:
        """
        Check if two objects are different.

        Args:
            other (Any): The other object to compare.
        """
        return not self.__eq__(other)

    def __or__(self, other: Any) -> Any:
        """
        This method is used to merge two objects using the `|` operator.
        It merges both objects using the `self.__dict__` method if `other` is an instance of the same class.
        Otherwise, if `other` is a dictionary, it merges `self.__dict__` with the `other` dictionary.

        Args:
            other (Any): The other object to merge.

        Returns:
            NotImplemented: If `other` is not an instance of the same class or an instance of a dictionary.
        """
        if isinstance(other, type(self)):
            return type(self)(**(self.__dict__ | other.__dict__))

        if isinstance(other, dict):
            return type(self)(**(self.__dict__ | other))

        # don't attempt to compare against unrelated types
        return NotImplemented

    def __repr__(self) -> str:
        """
        This method returns the representation of the object in a string format.

        Returns:
            str: The string representation of the object.
        """
        return self.__data__.__repr__()

    def __ror__(self, other: Any) -> Any:
        """
        This method is used to merge two objects using the `|` operator, with the current object being on the right-hand side.
        It merges the current object with `other` using the `__dict__` method if `other` is an instance of the same class.
        If `other` is a dictionary, it merges the `other` dictionary with `self.__dict__`.

        Args:
            other (Any): The other object to merge.

        Returns:
            NotImplemented: if `other` is not an instance of the same class or an instance of a dictionary.
        """
        if isinstance(other, type(self)):
            return type(self)(**(other.__dict__ | self.__dict__))

        if isinstance(other, dict):
            return type(self)(**(other | self.__dict__))

        # don't attempt to compare against unrelated types
        return NotImplemented

    def __setattr__(self, name: str, value: Any, caller: str = None) -> None:
        """
        Method changed from BaseClass for guarantee de integrity of data attributes.
        This method is executed on setting attributes in the object.
        Ex: obj.attr = 1

        Raises:
            AttributeError: If is frozen.
        """
        super().__setattr__(name, value)
        # Because self.__get_clean_value__ can change the value we need to pick it from self
        new_value = getattr(self, name)

        # When value is associated:
        # directly on the key Ex dict['key'] = value then caller will be __setitem__
        # directly on attribute Ex dict.key = value then caller will be None
        # Don't do that to private attributes
        if caller is None and self.is_valid_key(key=name):
            # For integrity guarantee writes the value to the dictionary key as well.
            self.__setitem__(name, new_value, caller='__setattr__')

    def __setitem__(self, key: str, item: Any, caller: str = None) -> None:
        """
        Method changed from BaseClass for guarantee de integrity of data keys.
        This method is executed on setting items in the dictionary.
        Ex: d = dict(key=1) or d['key'] = 1

        Raises:
            AttributeError: If is frozen.
        """
        if key.startswith('_'):
            raise KeyError("Keys can't start with '_'.")
        config = getattr(self, CONFIG_ATTRIBUTE_NAME)
        if key in config.keys_blacklist:
            raise KeyError(f'The key cannot be called "{key}".')

        # When value is associated:
        # directly on the key Ex dict['key'] = value then caller will be None
        # directly on attribute Ex dict.key = value then caller will be __setattr__
        if caller is None:
            # For integrity guarantee writes the value to the attribute as well.
            self.__setattr__(key, item, caller='__setitem__')

            # The setattr can "clean" the value them we need to catch it again
            item = getattr(self, key)

        # If key is a property we do not set on dict
        if not isinstance(getattr(type(self), key, None), property):
            self.__data__.__setitem__(key, item)

    def __setstate__(self, state: dict = None) -> None:
        """
        This method is used by Pickle module to set back the correct serialized data.
        We need to iterate over every key and set the value to the object.

        Args:
            state (dict): The result from the __getstate__ method used by Pickle.
        """
        # NOTE: coverage not passing here since `self` will always have __data__ attribute
        # For some old Pickle objects the data attribute will not exists so we need to create it
        if not hasattr(self, '__data__'):
            self.__data__ = {} # pylint: disable=attribute-defined-outside-init

        return super().__setstate__(state)


    ## Public Methods
    def clear(self) -> None:
        """
        This method clears the dictionary.

        Raises:
            AttributeError: If is frozen.
        """
        self.__check_frozen__()
        # Because we integrate key/attributes we need to remove booth
        # The original clear only remove keys
        # We need to convert keys to a list, because the original is a iterator
        for key in list(self.keys()):
            del self[key]

    def copy(self) -> dict:
        """
        Generate a copy for this object, we need to use deepcopy because
        this object could have some keys/attributes and they need to be on the copy.
        If the object is Frozen, the copy will not be.
        """
        return deepcopy(self)

    def fromkeys(self, keys: list, default: Any = None) -> dict:
        """
        Create a new object with the keys, if key does not exists add one with default value.
        If the object is Frozen, the copy will not be.

        Args:
            keys (list): The list with the keys for the new object.
            default (Any, optional): The default value if the key does not exists. Defaults to None.
        """
        # Create a new key list with all keys and use a set to avoid duplicates
        keys_aux = set(keys)
        keys_aux.update(self.keys())

        dct = self.copy()
        for key in keys_aux:
            if key in keys:
                dct[key] = dct.get(key, default)
            else:
                del dct[key]

        return dct

    def get(self, key: str, default: Any = None) -> Any:
        """
        Get the value for the key, if key does not exists return default.

        Args:
            key (str): The key to get the value.
            default (Any, optional): The default value if the key does not exists. Defaults to None.

        Returns:
            Any: The value for the key or the default value.
        """
        return self.__data__.get(key, default)

    def is_valid_key(self, key: str) -> bool:
        """
        This method checks if the key is valid.
        Valid keys are the ones that not starts with '_' and
        are not in the config.keys_blacklist.

        Args:
            key (str): The key that needs to be checked.

        Returns:
            bool: True if the key is valid, False otherwise.
        """
        config = getattr(self, CONFIG_ATTRIBUTE_NAME)
        return not key.startswith('_') and key not in config.keys_blacklist

    def items(self) -> dict_items:
        """
        Return a new view of the dictionary's items ((key, value) pairs).
        """
        return self.__data__.items()

    def keys(self) -> dict_keys:
        """
        Return a new view of the dictionary's keys.
        """
        return self.__data__.keys()

    def pop(self, *args) -> Any:
        """
        Remove specified key and return the corresponding value.
        If the key is not found return the default otherwise raise a KeyError.

        Args:
            key (str): The key that will be removed.
            default (Any, optional): The default value if key is not found.

        Raises:
            AttributeError: If is frozen.
            KeyError: If the default is not passed and key is not found.
        """
        self.__check_frozen__()
        ret = self.__data__.pop(*args)

        try:
            # Some times the attribute will not exists
            self.__delattr__(args[0]) # pylint: disable=unnecessary-dunder-call
        except AttributeError:
            pass

        return ret

    def popitem(self) -> tuple:
        """
        Remove and return a (key, value) pair as a 2-tuple.
        Pairs are returned in LIFO (last-in, first-out) order.

        Raises:
            AttributeError: If is frozen.
            KeyError: If the dict is empty.
        """
        self.__check_frozen__()
        return self.__data__.popitem()

    def update(self, *args, **kwargs) -> None:
        """
        Update self with the key/value pairs that are passed.
        We check every value to see if it is valid.

        Example:

            >>> dct = BaseDict()
            >>> dct.update({'a': 1})
            >>> dct
            {'a': 1}
            >>> dct.update([('b', 2)])
            >>> dct
            {'a': 1, 'b': 2}

        Raises:
            AttributeError: If is frozen.
        """
        self.__check_frozen__()
        dct = dict(*args, **kwargs)
        for key, value in dct.items():
            if self.is_valid_key(key=key):
                self[key] = value
            else:
                raise KeyError(f'The key cannot be called "{key}".')

    def values(self) -> dict_values:
        """
        Return a new view of the dictionary's values.
        """
        return self.__data__.values()



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import json

from typing import Any

from everysk.config import settings
from everysk.api import http_client
from everysk.core.exceptions import APIError

###############################################################################
#   APIRequestor Implementation
###############################################################################
class APIRequestor(object):

    def __init__(self, api_entry, api_version, api_sid, api_token, verify_ssl_certs) -> None:
        """
        Initializes an APIRequestor object with the necessary parameters to make HTTP requests to an API.

        This constructors sets up the APIRequestor with custom headers, including authorization, content type and user agent details.

        Args:
            api_entry (str): The base URL of the API endpoint.
            api_version (str): The current version of the API.
            api_sid (str): The API SID (identifier) for authentication.
            api_token (str): The API token for authentication.
            verify_ssl_certs (bool): Specified whether SSL certificates should be verified.

        Raises:
            Exception: If any of the parameters are invalid.

        Example:
            >>> from everysk.api.api_requestor import APIRequestor

            >>> api_requestor = APIRequestor(
            >>> ... api_entry='https://api.example.com',
            >>> ... api_version='v2',
            >>> ... api_sid='your_api_sid',
            >>> ... api_token='your_api_token',
            >>> ... verify_ssl_certs=True
            >>> ... )
        """
        if not api_entry:
            raise ValueError('Empty api_entry.') # pylint: disable=broad-exception-raised
        if api_version != 'v2':
            raise ValueError('Invalid api_version (supported version: "v2").') # pylint: disable=broad-exception-raised
        if not api_sid:
            raise ValueError('Invalid api_sid.') # pylint: disable=broad-exception-raised
        if not api_token:
            raise ValueError('Invalid api_token.') # pylint: disable=broad-exception-raised

        self.headers = settings.HTTP_DEFAULT_HEADERS.copy()
        self.headers.update(
            {
                'Content-Type': 'application/json',
                'Authorization': f'Bearer {api_sid}:{api_token}',
                'User-Agent': f'Everysk PythonBindings/{api_version}'
            }
        )
        self.base_url = f'{api_entry}/{api_version}'
        self.client = http_client.new_default_http_client(
            timeout=3600,
            verify_ssl_certs=verify_ssl_certs,
            allow_redirects=False
        )

    def _clean_response(self, code: int, response: str) -> Any:
        if code not in settings.HTTP_SUCCESS_STATUS_CODES:
            raise APIError(code, response)
        return json.loads(response)

    def get(self, path, params):
        """
        Sends a GET request to the specified path with the given parameters.

        Args:
            path (str): The path to send the GET request to.
            params (dict): The parameters to include in the GET request.

        Returns:
            dict: The JSON response from the GET request.

        Raises:
            APIError: If the response code is not in the HTTP_SUCCESS_STATUS_CODES list.
        """
        url = f'{self.base_url}{path}'
        code, response = self.client.request(
            'GET',
            url,
            headers=self.headers,
            params=params
        )
        return self._clean_response(code, response)

    def post(self, path, payload):
        """
        Sends a POST request to the specified path with the given payload.

        Args:
            path (str): The path to send the request to.
            payload (dict): The payload to include in the request.

        Raises:
            APIError: If the response code is not in the HTTP_SUCCESS_STATUS_CODES list.

        Returns:
            dict: The JSON response from the API.
        """
        url = f'{self.base_url}{path}'
        code, response = self.client.request(
            'POST',
            url,
            headers=self.headers,
            payload=payload
        )
        return self._clean_response(code, response)

    def delete(self, path):
        """
        Sends a DELETE request to the specified path.

        Args:
            path (str): The path to send the DELETE request to.

        Raises:
            APIError: If the response code is not in the HTTP_SUCCESS_STATUS_CODES list.

        Returns:
            dict: The JSON response from the server.
        """
        url = f'{self.base_url}{path}'
        code, response = self.client.request(
            'DELETE',
            url,
            headers=self.headers
        )
        return self._clean_response(code, response)

    def put(self, path, payload):
        """
        Sends a PUT request to the specified path with the given payload.

        Args:
            path (str): The path to send the request to.
            payload (dict): The payload to include in the request.

        Raises:
            APIError: If the response code is not in the HTTP_SUCCESS_STATUS_CODES list.

        Returns:
            dict: The JSON response from the server.
        """
        url = f'{self.base_url}{path}'
        code, response = self.client.request(
            'PUT',
            url,
            headers=self.headers,
            payload=payload
        )
        return self._clean_response(code, response)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api.api_resources.api_resource import (
    RetrievableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
    CreateableAPIResource,
    UpdateableAPIResource
)

###############################################################################
#   File Implementation
###############################################################################
class File(
    RetrievableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
    CreateableAPIResource,
    UpdateableAPIResource
):
    """
    Represents a file resource.

    This class inherits from the following API resource classes:
        - RetrievableAPIResource
        - ListableAPIResource
        - DeletableAPIResource
        - CreateableAPIResource
        - UpdateableAPIResource
    """
    pass



# -*- coding: utf_8 -*-
###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import Any, TYPE_CHECKING
import csv
import six

from everysk.config import settings
from everysk.core.fields import ListField, StrField, FloatField, BoolField
from everysk.core.string import to_string
from everysk.sdk.entities.base import BaseEntity, ScriptMetaClass
from everysk.sdk.entities.fields import EntityNameField, EntityDescriptionField, EntityLinkUIDField, EntityWorkspaceField, EntityDateTimeField, EntityTagsField, CurrencyField
from everysk.sdk.entities.portfolio.securities import Securities
from everysk.sdk.entities.script import Script

if TYPE_CHECKING:
    from io import StringIO


###############################################################################
#   SecuritiesField Class Implementation
###############################################################################
class SecuritiesField(ListField):
    attr_type = Securities

    def __init__(self, default: Any = None, required_lazy: bool = True, **kwargs) -> None:
        """
        This method initializes the field.

        Args:
            default (Any): The default value of the field.
            required_lazy (bool): True if the field is required, False otherwise.
            **kwargs (dict): The keyword arguments of the field.
        """
        super().__init__(default=default, required_lazy=required_lazy, **kwargs)

    def clean_value(self, value: Any) -> Securities:
        """
        This method cleans the value of the field.

        Args:s
            value (Any): The value to be cleaned.

        Returns:
            Securities: The cleaned value.
        """
        if isinstance(value, list):
            value = self.attr_type(value)
        return super().clean_value(value)


###############################################################################
#   Portfolio Class Implementation
###############################################################################
class Portfolio(BaseEntity, metaclass=ScriptMetaClass):
    """
    This class represents a portfolio entity object and provides methods to validate
    and convert the entity object to a dictionary or a CSV string representation.

    Attributes:
        script (Script): The script object associated to the entity.
        id (StrField): The unique identifier of the portfolio.
        workspace (EntityWorkspaceField): The workspace of the portfolio.
        name (EntityNameField): The name of the portfolio.
        tags (EntityTagsField): The tags of the portfolio.
        link_uid (EntityLinkUIDField): The link uid of the portfolio.
        description (EntityDescriptionField): The description of the portfolio.
        nlv (FloatField): The nlv of the portfolio.
        base_currency (CurrencyField): The base currency of the portfolio.
        date (EntityDateTimeField): The date of the portfolio.
        securities (SecuritiesField): The securities of the portfolio.
        level (StrField): The level of the portfolio.

        version (StrField): The version of the portfolio.
        created_on (DateTimeField): The created on date of the portfolio.
        updated_on (DateTimeField): The updated on date of the portfolio.

    Example:
        >>> from everysk.sdk.entities.portfolio.base import Portfolio
        >>> portfolio = Portfolio()
        >>> portfolio.workspace = 'my_workspace'
        >>> portfolio.name = 'My Portfolio'
        >>> portfolio.tags = ['tag1', 'tag2']
        >>> portfolio.nlv = 1000.0
        >>> portfolio.base_currency = 'USD'
        >>> portfolio.date = '20210101'
        >>> portfolio.securities = ['sec1', 'sec2', 'sec3']
        >>> portfolio.save()
        >>> print(portfolio)
        {
            'id': 'port_12345678',
            'workspace': 'my_workspace',
            'name': 'My Portfolio',
            'tags': ['tag1', 'tag2'],
            'nlv': 1000.0,
            'base_currency': 'USD',
            'date': '20210101',
            'securities': ['sec1', 'sec2', 'sec3'],
            'version': 'v1',
            'created_on': '2021-01-01T00:00:00.000000Z',
            'updated_on': '2021-01-01T00:00:00.000000Z',
            'level': 'portfolio'
        }

    """
    script: Script
    _orderable_attributes = ListField(default=['date', 'created_on', 'updated_on', 'name'], readonly=True)

    id = StrField(regex=settings.PORTFOLIO_ID_REGEX, max_size=settings.PORTFOLIO_ID_MAX_SIZE, required_lazy=True, empty_is_none=True)

    name = EntityNameField()
    description = EntityDescriptionField()
    tags = EntityTagsField()
    link_uid = EntityLinkUIDField()
    workspace = EntityWorkspaceField()

    nlv = FloatField(min_size=float('-inf'), max_size=float('inf'))
    base_currency = CurrencyField()
    date = EntityDateTimeField()
    securities = SecuritiesField(min_size=0, max_size=settings.PORTFOLIO_MAX_SIZE, required_lazy=True)
    level = StrField(min_size=1, max_size=settings.PORTFOLIO_LEVEL_MAX_LENGTH)

    check_securities = BoolField(default=False)

    ###############################################################################
    #   Portfolio Basics Implementation
    ###############################################################################

    @staticmethod
    def get_id_prefix() -> str:
        """
        Returns the prefix of the portfolio id field value.

        Returns:
            str: The prefix of the portfolio id field value.

        Example:
            >>> Portfolio.get_id_prefix()
            'port_'

        Notes:
            The prefix is typically used to distinguish portfolio IDs from other types of IDs.
        """
        return settings.PORTFOLIO_ID_PREFIX

    def validate(self) -> bool:
        """
        This method validates the entity object and raises an exception if it is not
        valid. The validation is performed by calling the `validate` method of each field
        of the entity.

        Args:
            self (Self): The entity object to be validated.

        Raises:
            FieldValueError: If the entity object is not valid.
            RequiredFieldError: If a required field is missing.

        Returns:
            bool: True if the entity object is valid, False otherwise.

        Example:
            >>> entity = Entity()
            >>> entity.validate()
            True
        """
        return self.get_response(self_obj=self)

    def to_csv(self) -> str:
        """
        This method converts the entity object into a CSV string representation of the
        entity object.

        Returns:
            str: A CSV string representation of the entity.

        Example:
            >>> entity = Entity()
            >>> entity.to_csv()
            "name,date,base_currency,nlv,description,tags,securities
            My Portfolio,2021-01-01,USD,1000.0,My Portfolio Description,tag1 tag2,sec1,sec2,sec3"
        """
        def write_row(csv_writer_: Any, line: list[str], length: int) -> None:
            """
            Write a row to a CSV file

            Args:
                csv_writer_ (Any): CSV writer object to use for writing the row
                line (list[str]): The list of values to write as a row
                length (int): The desired length of the row

            Returns:
                None

            Example:
                >>> from everysk.sdk.entities.portfolio.base import Portfolio
                >>> import csv
                >>> csv_file = open('output'.csv', 'w', newline='')
                >>> write = csv.writer(csv_file)
                >>> Portfolio.write_row(writer, ['value1', 'value2'], 5)
                >>> csv_file.close()

            Notes:
                This function pads the row with empty strings to ensure the fixed length.
            """
            out: list[str] = line + ([''] * (length - len(line)))
            out = [to_string(s) if s is not None else None for s in out]
            csv_writer_.writerow(out)
            return None

        csv_file: StringIO = six.StringIO()
        csv_writer: Any = csv.writer(csv_file)

        date_: str = self.date.strftime()
        nlv_: str = '' if not self.nlv else to_string(self.nlv)
        pre_headers: list[str] = [self.name, date_, self.base_currency, nlv_, self.description, ' '.join(self.tags)]

        securities_as_lists: list[list[str]] = self.securities.to_lists()

        headers: list[str] = securities_as_lists[0]
        headers_len: int = len(headers)
        write_row(csv_writer, pre_headers, headers_len)
        write_row(csv_writer, headers, headers_len)

        for sec_as_list in securities_as_lists[1:]:
            write_row(csv_writer, sec_as_list, headers_len)

        out: str = csv_file.getvalue()
        csv_file.close()
        return out

    def to_dict(self, add_class_path: bool = False, recursion: bool = False) -> dict:
        """
        This method is used to convert the object to a dictionary.
        """
        dct: dict = super().to_dict(add_class_path=add_class_path, recursion=recursion)

        if isinstance(dct['securities'], Securities):
            dct['securities'] = dct['securities'].to_list(add_class_path=add_class_path, recursion=recursion)

        return dct



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from types import NoneType
from typing import Any, Self
from collections.abc import Iterable

from six import string_types

from everysk.config import settings
from everysk.core.datetime import DateTime, Date
from everysk.core.exceptions import FieldValueError
from everysk.core.number import is_float_convertible
from everysk.core.string import normalize_string, to_string
from everysk.sdk.entities.base_list import EntityList


###############################################################################
#   Tags Class Implementation
###############################################################################
class Tags(EntityList):
    """
    A specialized list for handling tags.

    This class is a subclass of EntityList and provides specific validation for tags, including
    size limits and pattern matching.

    Attributes:
        min_size (int): The minimum allowed size for the list (default is 1).
        max_size (int): The maximum allowed size for the list (default is the value from settings).

    Example:
        To create a Tags for handling tags:
        tags = Tags()
    """
    min_size: int = 1
    max_size: int = settings.ENTITY_MAX_TAG_LENGTH

    def __init__(self, *args):
        """
        Initialize the Tags.

        Args:
            *args: Optional initial elements to populate the list.

        Example:
            To create an Tags with initial elements:
            >>> my_list = Tags(['tag1', 'tag2', 'tag3'])
        """
        if args:
            try:
                args = (list(dict.fromkeys(self.unify(args[0]))), )
            except TypeError:
                raise FieldValueError('Unsupported format in Tags.') from TypeError
        super().__init__(*args)

    def _validate(self, value: Any) -> str:
        """
        Validate a tag value.

        Args:
            value (str): The tag value to be validated.

        Raises:
            FieldValueError: If the tag value is invalid based on the specified criteria.

        Returns:
            str: The validated tag value.

        Criteria:
            - The value must be a string.
            - The value's length must be between min_size ({min_size}) and max_size ({max_size}).
            - The value must consist of only lowercase letters, digits, and underscores.
            - The value cannot be empty.

        Raises:
            FieldValueError: If the tag value does not meet the specified criteria.
        """
        if isinstance(value, list) or isinstance(value, dict):
            raise FieldValueError(f'Unsupported format in Tags: {type(value)}')
        elif value == '' or value is Undefined:
            value = str(None)
        elif is_float_convertible(value) or value.__class__ in [NoneType, bool]:
            value = str(value)
        elif value.__class__ in [DateTime, Date]:
            value = value.strftime('%Y%m%d')
        elif not isinstance(value, string_types):
            raise FieldValueError(f'Unsupported format in Tags: {type(value)}')

        value = to_string(''.join(x if x.isalnum() else '_' for x in normalize_string(value))).lower()

        if (
            self.min_size is not None and self.min_size > len(value) or
            self.max_size is not None and self.max_size < len(value)
        ):
            raise FieldValueError(f"Tags: '{value}' size it's not between {self.min_size} and {self.max_size}")

        return value

    def insert(self, __index: int, __object: Any) -> None:
        """
        Insert an object at a specified index if the object does not exist with validation.

        Args:
            __index (int): The index at which to insert the object.
            __object: The object to be inserted.
        """
        __object = self._validate(__object)
        if __object not in self:
            super().insert(__index, __object)

    def append(self, __object: Any) -> None:
        """
        Append an object to the list if the object does not exist with validation.

        Args:
            __object: The object to be appended.
        """
        __object = self._validate(__object)
        if __object not in self:
            super().append(__object)

    def extend(self, __iterable: Iterable) -> None:
        """
        Extend the list with values from an iterable if the value does not exist with validation.

        Args:
            __iterable (Iterable): An iterable containing values to be added to the list.
        """
        order_values: dict = {}
        for value in __iterable:
            value: str = self._validate(value)
            if value not in self:
                order_values[value] = None
        super().extend(list(order_values.keys()))

    @classmethod
    def unify(cls, tag_set: Any, reference_tags: list[Any] = None) -> Self:
        """
        Unify a set of tags with a reference set of tags.

        Args:
            tag_set (list): The set of tags to be unified.
            reference_tags (list): The reference set of tags.

        Returns:
            list: The unified set of tags.

        Example:
            To unify a set of tags with a reference set of tags:
            >>> unified_tags = Tags.unify_tags([['tag1', 'tag2'], ['tag3', 'tag4']], ['tag5', 'tag6'])
        """
        unified_tags: Self = cls()

        if not tag_set:
            return unified_tags
        elif isinstance(tag_set, (int, float)):
            tag_set = [tag_set]

        for tags in tag_set:
            if tags is not False and not tags and reference_tags:
                unified_tags.extend(reference_tags)
            elif isinstance(tags, list):
                unified_tags.extend(cls.unify(tags, reference_tags))
            else:
                unified_tags.append(tags)

        return unified_tags



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api.api_resources import Report
from everysk.core.unittests import TestCase, mock

###############################################################################
#   Report TestCase Implementation
###############################################################################
class APIReportTestCase(TestCase):

    def test_class_name_method(self):
        self.assertEqual(Report.class_name(), 'report')

    @mock.patch('everysk.api.utils.create_api_requestor')
    def test_share_method(self, mock_create_api_requestor):
        mock_api_req = mock.Mock()
        mock_create_api_requestor.return_value = mock_api_req

        mock_response_data = {'id': 'report_id', 'other_data': 'example'}
        mock_api_req.post.return_value = {'report': mock_response_data}
        report = Report(retrieve_params={'key1': 'value1'}, params={'key2': 'value2'})
        report['id'] = 'report_id'
        result = report.share(param1='value1', param2='value2')
        expected_url = '/reports/report_id/share'

        mock_api_req.post.assert_called_once_with(expected_url, {'param1': 'value1', 'param2': 'value2'})
        self.assertEqual(report['id'], 'report_id')
        self.assertEqual(report['other_data'], 'example')
        self.assertNotIn('param1', report)
        self.assertNotIn('param2', report)
        self.assertEqual(result, report)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
__all__ = ['Route', 'RouteLazy']

from starlette.routing import Route, compile_path
from starlette.types import Receive, Scope, Send

from everysk.core.string import import_from_string


###############################################################################
#   RouteLazy Class Implementation
###############################################################################
class RouteLazy(Route):
    app: callable = None
    endpoint: str = None
    include_in_schema: bool = True
    methods: set = None
    name: str = None
    path: str = None

    def __init__(self, path: str, endpoint: str, name: str | None = None) -> None: # pylint: disable=super-init-not-called
        """
        Route class that will lazy load the endpoint class.

        Args:
            path (str): The path of the route always start with '/'.
            endpoint (str): Full doted class path of the endpoint.
            name (str | None, optional): A name for this endpoint. Defaults to the same value in endpoint.

        Raises:
            ValueError: If the path doesn't start with '/'.
        """
        if not path.startswith('/'):
            raise ValueError("Routed paths must start with '/'")

        self.path = path
        self.endpoint = endpoint
        self.name = endpoint if name is None else name
        self.path_regex, self.path_format, self.param_convertors = compile_path(path)

    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Method to handle the incoming request and execute it.
        If the endpoint is a string, it will import the endpoint class and execute it.

        Args:
            scope (Scope): ASGI scope dictionary.
            receive (Receive): ASGI receive data.
            send (Send): ASGI send data.
        """
        if isinstance(self.endpoint, str):
            self.app = import_from_string(self.endpoint)
        else:
            self.app = self.endpoint

        await self.app(scope, receive, send)



# -*- coding: utf_8 -*-
###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import Any

from everysk.config import settings
from everysk.core.log import Logger
from everysk.core.datetime import DateTime
from everysk.core.fields import ChoiceField, StrField, DateTimeField, ListField, _min_max_validate
from everysk.sdk.engines.market_data import MarketData
from everysk.sdk.entities.tags import Tags


log = Logger(name=__name__)


###############################################################################
#   CurrencyField Class Implementation
###############################################################################
class CurrencyField(ChoiceField):
    """
    A field for currency codes with validation capabilities.

    This class extends the standard ChoiceField and adds validation for the values
    added or inserted into it.
    """
    _choices: list[str] = None
    _market_data: MarketData = MarketData()

    def __init__(self, default: str = None, choices: list = None, readonly: bool = False, required: bool = False, required_lazy: bool = True, empty_is_none: bool = False, **kwargs) -> None:
        super().__init__(default=default, choices=choices, readonly=readonly, required=required, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)

    @property
    def choices(self) -> list[str]:
        """
        Get the available choices for the field.

        Returns:
            A list of available choices.
        """
        if not self._choices:
            self._choices = self._get_base_currencies()
        return self._choices

    @choices.setter
    def choices(self, value: list[str]) -> None:
        """
        Set the available choices for the field.

        Args:
            value (list[str]): The available choices.
        """
        self._choices = value

    def _get_base_currencies(self) -> list[str]:
        """
        Get all available currencies.

        Returns:
            A list of all available currencies.
        """
        try:
            currencies = self._market_data.get_currencies()
        except Exception: # pylint: disable=broad-except
            log.warning("Failed to get currencies from MarketData.")
            currencies = None

        if currencies is None:
            return list(settings.ENTITY_BASE_CURRENCY_DEFAULT_LIST) + [None]

        return list(dict(currencies['base_currencies']).keys()) + [None]


###############################################################################
#   EntityNameField Class Implementation
###############################################################################
class EntityNameField(StrField):
    """
    This class is a subclass of StrField and provides specific validation for name, including
    size limits and pattern matching.

    Attributes:
        min_size (int): The minimum allowed size for the list (default is the value from settings).
        max_size (int): The maximum allowed size for the list (default is the value from settings).
    """
    attr_type = str

    def __init__(
        self,
        default: Any = None,
        regex: str = None,
        min_size: int = settings.ENTITY_NAME_MIN_LENGTH,
        max_size: int = settings.ENTITY_NAME_MAX_LENGTH,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = True,
        empty_is_none: bool = True,
        **kwargs
    ) -> None:
        super().__init__(default=default, regex=regex, min_size=min_size, max_size=max_size, readonly=readonly, required=required, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)


###############################################################################
#   EntityDescriptionField Class Implementation
###############################################################################
class EntityDescriptionField(StrField):
    """
    This class is a subclass of StrField and provides specific validation for description, including
    size limits and pattern matching.

    Attributes:
        min_size (int): The minimum allowed size for the list (default is the value from settings).
        max_size (int): The maximum allowed size for the list (default is the value from settings).
    """
    attr_type = str

    def __init__(
        self,
        default: Any = None,
        regex: str = None,
        min_size: int = settings.ENTITY_DESCRIPTION_MIN_LEN,
        max_size: int = settings.ENTITY_DESCRIPTION_MAX_LEN,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        super().__init__(default=default, regex=regex, min_size=min_size, max_size=max_size, readonly=readonly, required=required, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)


###############################################################################
#   EntityLinkUIDField Class Implementation
###############################################################################
class EntityLinkUIDField(StrField):
    """
    This class is a subclass of StrField and provides specific validation for link uid, including
    size limits and pattern matching.

    Attributes:
        min_size (int): The minimum allowed size for the list (default is the value from settings).
        max_size (int): The maximum allowed size for the list (default is the value from settings).
    """
    attr_type = str

    def __init__(
        self,
        default: Any = None,
        regex: str = None,
        min_size: int = settings.ENTITY_LINK_UID_MIN_LENGTH,
        max_size: int = settings.ENTITY_LINK_UID_MAX_LENGTH,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = True,
        **kwargs
    ) -> None:
        super().__init__(default=default, regex=regex, min_size=min_size, max_size=max_size, readonly=readonly, required=required, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)


###############################################################################
#   EntityWorkspaceField Class Implementation
###############################################################################
class EntityWorkspaceField(StrField):
    """
    This class is a subclass of StrField and provides specific validation for workspace, including
    size limits and pattern matching.

    Attributes:
        min_size (int): The minimum allowed size for the list (default is the value from settings).
        max_size (int): The maximum allowed size for the list (default is the value from settings).
    """
    attr_type = str

    def __init__(
        self,
        default: Any = None,
        regex: str = settings.ENTITY_WORKSPACE_REGEX,
        min_size: int = settings.ENTITY_WORKSPACE_MIN_LENGTH,
        max_size: int = settings.ENTITY_WORKSPACE_MAX_LENGTH,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = True,
        empty_is_none: bool = True,
        **kwargs
    ) -> None:
        super().__init__(default=default, regex=regex, min_size=min_size, max_size=max_size, readonly=readonly, required=required, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)


###############################################################################
#   EntityDateTimeField Class Implementation
###############################################################################
class EntityDateTimeField(DateTimeField):
    """
    This class is a subclass of StrField and provides specific validation for date, including
    size limits and pattern matching.

    Attributes:
        min_size (int): The minimum allowed size for the list (default is the value from DateTime market start).
        max_size (int): The maximum allowed size for the list (default is one day delta from now).
    """
    attr_type = DateTime

    def __init__(
        self,
        default: Any = None,
        min_date: DateTime = None,
        max_date: DateTime = None,
        force_time: str = 'MIDDAY',
        required: bool = False,
        readonly: bool = False,
        required_lazy: bool = True,
        empty_is_none: bool = True,
        **kwargs
    ) -> None:
        super().__init__(default, min_date=min_date, max_date=max_date, force_time=force_time, required=required, readonly=readonly, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)

    def validate(self, attr_name: str, value: Any, attr_type: type = None) -> None:
        """
        Checks if value is greater than min and lower than max including both values.

        Args:
            attr_name (str): The name of the attribute being validated.
            value (Any): The value to validate.
            attr_type (type): The type of the attribute being validated.

        Raises:
            FieldValueError: If the value is not within the specified range.

        Example:
            >>> from everysk.sdk.entities.fields import EntityDateTimeField
            >>> from everysk.core.datetime.datetime import DateTime
            >>> field = EntityDateTimeField(min_date=DateTime(2023-01-01), max_date=DateTime(2023-12-31))
            >>> try:
            >>> ... field.validate("test_field", DateTime(2023, 6, 15))
            >>> ... print("Valid Date")
            >>> except Exception as e:
            >>> ... print(f"Validation error: {e}")
            >>> Valid Date
        """
        min_date = self.min_date if self.min_date is not None else DateTime.market_start()
        max_date = self.max_date if self.max_date is not None else DateTime.now().delta(1, 'D').force_time('LAST_MINUTE')
        _min_max_validate(min_date, max_date, value, attr_name)
        return super().validate(attr_name, value, attr_type)


###############################################################################
#   EntityTagsField Class Implementation
###############################################################################
class EntityTagsField(ListField):
    """
    This class is a subclass of ListField and provides specific validation for tags, including
    size limits and pattern matching.

    Attributes:
        min_size (int): The minimum allowed size for the list (default is the value from settings).
        max_size (int): The maximum allowed size for the list (default is the value from settings).
    """
    attr_type = Tags

    def __init__(
        self,
        default: Any = None,
        min_size: int = settings.ENTITY_MIN_TAG_SIZE,
        max_size: int = settings.ENTITY_MAX_TAG_SIZE,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        if default is not None and not isinstance(default, Tags):
            default = Tags(default)
        super().__init__(default=default, min_size=min_size, max_size=max_size, readonly=readonly, required=required, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)

    def clean_value(self, value: Any) -> Any:
        """
        This method ensures that the provided value is in the expected format before assigning it to an attribute.
        If the value is None, it is replaced with an empty TagsList. If it is not already a TagsList instance,
        it is converted into one.

        Args:
            value (Any): The value to clean, None, A TagsList instance, or any other value.

        Example:
            >>> from everysk.sdk.entities.tags import Tags
            >>> from everysk.sdk.entities.fields import EntityTagsField
            >>> field = EntityTagsField()

            >>> cleaned_value_none = field.clean_value(None)
            >>> print(cleaned_value_none)
            >>> []

            >>> cleaned_non_empty_tags = field.clean_value(Tags(['tag1', 'tag2']))
            >>> print(cleaned_non_empty_tags)
            >>> ['tag1', 'tag2']
        """
        value = super().clean_value(value)
        if value is None:
            value = Tags()
        elif not isinstance(value, Tags):
            value = Tags(value)
        return value



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.datetime import Date, DateTime
from everysk.core.exceptions import FieldValueError
from everysk.core.object import BaseDict
from everysk.core.unittests import TestCase

from everysk.sdk.entities.portfolio.security import Security


###############################################################################
#   Security TestCase Implementation
###############################################################################
class TestSecurity(TestCase):

    def setUp(self):
        # Sample data for valid initialization
        self.valid_data = {
            'status': 'OK',
            'symbol': 'AAPL',
            'id': 'unique123',
            'quantity': 100.0,
            'instrument_class': 'Equity',
            'maturity_date': Date(2025, 5, 1),
            'issue_date': Date(2020, 1, 1),
            'return_date': Date(2025, 12, 31),
            'settlement': Date(2020, 1, 2),
            'display': 'Apple Inc.',
            'comparable': 'Yes',
            'previous_quantity': 50.0,
            'extra_field1': 'extra_value1',  # This should be placed in `extra_data`
            'extra_field2': 'extra_value2'   # This should also be placed in `extra_data`
        }
        self.security = Security(**self.valid_data)

    def test_initialization(self):
        base_security = {
            'return_date': None,
            'status': None,
            'label': None,
            'ticker': None,
            'underlying': None,
            'symbol': None,
            'coupon': None,
            'issuer': None,
            'cost_price': None,
            'book': None,
            'warranty': None,
            'instrument_class': None,
            'maturity_date': None,
            'percent_index': None,
            'issuer_type': None,
            'fx_rate': None,
            'operation': None,
            'market_value_in_base': None,
            'asset_subclass': None,
            'id': None,
            'exchange': None,
            'premium': None,
            'instrument_type': None,
            'isin': None,
            'indexer': None,
            'unrealized_pl_in_base': None,
            'accounting': None,
            'hash': None,
            'comparable': None,
            'error_message': None,
            'asset_class': None,
            'issue_date': None,
            'display': None,
            'market_price': None,
            'multiplier': None,
            'unrealized_pl': None,
            'series': None,
            'trader': None,
            'name': None,
            'error_type': None,
            'strike': None,
            'settlement': None,
            'option_type': None,
            'market_value': None,
            'previous_quantity': None,
            'quantity': None,
            'trade_id': None,
            'issue_price': None,
            'rate': None,
            'instrument_subtype': None,
            'extra_data': None,
            'currency': None,
            'look_through_reference': None
        }

        self.assertDictEqual(Security(), base_security)
        cmp_dict = base_security.copy()
        cmp_dict.update(extra_data={'var1': 2, 'var4': Date(2023, 1, 2), 'var5': DateTime(2023, 1, 2, 2, 3, 4), 'var6': None})

        self.assertDictEqual(
            Security(
                extra_data={
                    'var1': 1,
                    'var2': [1,2,3],
                    'var3': {'a': 1},
                    'var4': Date(2023, 1, 2),
                    'var5': DateTime(2023, 1, 1, 2, 3, 4),
                    'var6': Undefined
                },
                var1=2,
                var2=[1,2,3],
                var3={'a': 1},
                var4= Date(2023, 1, 2),
                var5=DateTime(2023, 1, 2, 2, 3, 4),
                var6=None
            ),
            cmp_dict
        )

    def test_convert_extra_data(self):
        extra_data = {'var1': 1}
        security = Security()
        security.extra_data = extra_data
        self.assertIsInstance(security.extra_data, BaseDict)
        self.assertEqual(security.extra_data, BaseDict(**extra_data))

    def test_initialization_with_valid_data(self):
        security = Security(**self.valid_data)
        self.assertEqual(security.status, 'OK')
        self.assertEqual(security.id, 'unique123')
        self.assertEqual(security.extra_data['extra_field1'], 'extra_value1')
        self.assertEqual(security.extra_data['extra_field2'], 'extra_value2')

    def test_from_list_conversion(self):
        headers = list(self.valid_data.keys())
        values = list(self.valid_data.values())
        security = Security.from_list(values, headers)
        self.assertEqual(security.status, 'OK')
        self.assertEqual(security.id, 'unique123')
        self.assertEqual(security.extra_data['extra_field1'], 'extra_value1')

    def test_initialization_with_extra_data(self):
        data_with_extra = self.valid_data.copy()
        data_with_extra['random_key'] = 'random_value'
        security = Security(**data_with_extra)
        self.assertEqual(security.extra_data['random_key'], 'random_value')

    def test_symbol_length_validation(self):
        invalid_data = self.valid_data.copy()
        invalid_data['symbol'] = 'A' * (settings.SYMBOL_ID_MAX_LEN + 1)
        with self.assertRaisesRegex(FieldValueError, "The length '101' for attribute 'symbol' must be between '0' and '100'."):
            Security(**invalid_data)

    def test_security_id_length_validation(self):
        invalid_data = self.valid_data.copy()
        invalid_data['id'] = '1' * (settings.SYMBOL_ID_MAX_LEN + 1)
        with self.assertRaisesRegex(FieldValueError, "The length '101' for attribute 'id' must be between '1' and '100'."):
            Security(**invalid_data)

    def test_extra_data_initialization(self):
        # Test that fields not in the class annotations are moved to `extra_data`
        data_with_extra = self.valid_data.copy()
        data_with_extra['extra_data'] = {'random_field': 'random_value'}
        security = Security(**data_with_extra)
        self.assertIn('random_field', security.extra_data)
        self.assertEqual(security.extra_data['random_field'], 'random_value')


    def test_extra_data_initialization_with_wrong_type(self):
        # Test if extra_data is an instance of BaseDict or dict
        with self.assertRaises(FieldValueError) as context:
            data_with_extra = self.valid_data.copy()
            data_with_extra['extra_data'] = 'invalid extra data'
            Security(**data_with_extra)

        self.assertEqual(str(context.exception), f'attribute extra_data must be a dictionary. {str}')

    def test_extra_data_empty_after_initialization(self):
        # Test that if no extra data is present, `extra_data` is set to None
        data_without_extra = self.valid_data.copy()
        data_without_extra.pop('extra_field1')
        data_without_extra.pop('extra_field2')
        security = Security(**data_without_extra)
        self.assertIsNone(security.extra_data)

    def test_generate_security_id(self):
        security_id1 = Security.generate_security_id()
        security_id2 = Security.generate_security_id()

        self.assertIsInstance(security_id1, str)
        self.assertIsInstance(security_id2, str)
        self.assertNotEqual(security_id1, security_id2)

    def test_sort_header(self):
        headers = ['status', 'symbol', 'quantity', 'instrument_class']
        headers_cmp = ['quantity', 'instrument_class', 'status', 'symbol']
        sorted_headers = Security.sort_header(headers_cmp)
        self.assertListEqual(sorted_headers, headers)

    def test_validate_required_fields_without_id(self):
        # Test if ID gets auto-generated if not provided
        data_without_id = self.valid_data.copy()
        data_without_id.pop('id')
        security = Security(**data_without_id)
        security.validate_required_fields()
        self.assertIsNotNone(security.id)

    def test_validate_required_fields_with_id(self):
        # Test if provided ID is not changed
        data_with_id = self.valid_data.copy()
        data_with_id['id'] = 'test_id'
        security = Security(**data_with_id)
        security.validate_required_fields()
        self.assertEqual(security.id, "test_id")

    def test_get_attr(self):
        ret = Security._get_attr(self.security, 'symbol') # pylint: disable=protected-access
        self.assertEqual(ret, 'AAPL')

    def test_get_attr_in_extra_data(self):
        valid_data = self.valid_data.copy()
        valid_data['extra_data'] = {'other_key': 'other_value'}
        security = Security(**valid_data)
        ret = Security._get_attr(security, 'other_key') # pylint: disable=protected-access
        self.assertEqual(ret, 'other_value')

    def test_get_attr_fallback(self):
        ret = Security._get_attr(self.security, 'inexistent_key', Security.generate_security_id) # pylint: disable=protected-access
        self.assertIsInstance(ret, str)

    def test_get_attr_inexistent(self):
        ret = Security._get_attr(self.security, 'inexistent_key') # pylint: disable=protected-access
        self.assertIsNone(ret)

    def test_generate_consolidation_key(self):
        ret = self.security.generate_consolidation_key(['symbol', 'instrument_class'])
        self.assertEqual(ret, 'AAPL_Equity')

    def test_to_list_without_headers(self):
        ret = self.security.to_list()
        self.assertListEqual(ret, ['OK', 'unique123', 'AAPL', 100.0, 'Equity', None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, '20250501', None, None, None, None, None, None, None, None, None, None, '20200101', None, None, None, None, None, None, None, None, None, None, None, '20251231', '20200102', None, {'extra_field1': 'extra_value1', 'extra_field2': 'extra_value2'}, None, 'Yes', 'Apple Inc.', None, None, 50.0])

    def test_to_dict_security(self):
        result = self.security.to_dict()
        expected_data = {
            'accounting': None,
            'instrument_subtype': None,
            'cost_price': None,
            'series': None,
            'operation': None,
            'trade_id': None,
            'issuer_type': None,
            'market_price': None,
            'ticker': None,
            'indexer': None,
            'exchange': None,
            'asset_subclass': None,
            'unrealized_pl': None,
            'display': 'Apple Inc.',
            'status': 'OK',
            'coupon': None,
            'issuer': None,
            'symbol': 'AAPL',
            'book': None,
            'currency': None,
            'market_value_in_base': None,
            'trader': None,
            'label': None,
            'rate': None,
            'warranty': None,
            'fx_rate': None,
            'percent_index': None,
            'comparable': 'Yes',
            'error_type': None,
            'option_type': None,
            'isin': None,
            'market_value': None,
            'issue_date': '20200101',
            'name': None,
            'instrument_type': None,
            'quantity': 100.0,
            'error_message': None,
            'maturity_date': '20250501',
            'underlying': None,
            'strike': None,
            'issue_price': None,
            'previous_quantity': 50.0,
            'return_date': '20251231',
            'settlement': '20200102',
            'id': 'unique123',
            'extra_data': {'extra_field1': 'extra_value1', 'extra_field2': 'extra_value2'},
            'look_through_reference': None,
            'instrument_class': 'Equity',
            'multiplier': None,
            'asset_class': None,
            'unrealized_pl_in_base': None,
            'hash': None,
            'premium': None
        }

        self.assertIsInstance(result, dict)
        self.assertEqual(expected_data, result)

    def test_to_dict_with_internals_attributes(self):
        result = self.security.to_dict(add_class_path=True)
        expected_data = {
            'warranty': None,
            'currency': None,
            'operation': None,
            'market_price': None,
            'premium': None,
            'instrument_type': None,
            'status': 'OK',
            'cost_price': None,
            'error_type': None,
            'label': None,
            'unrealized_pl_in_base': None,
            'id': 'unique123',
            'instrument_subtype': None,
            'look_through_reference': None,
            'unrealized_pl': None,
            'multiplier': None,
            'issue_price': None,
            'comparable': 'Yes',
            'issue_date': '20200101',
            'display': 'Apple Inc.',
            'asset_class': None,
            'name': None,
            'rate': None,
            'issuer_type': None,
            'hash': None,
            'indexer': None,
            'settlement': '20200102',
            'accounting': None,
            'coupon': None,
            'asset_subclass': None,
            'trade_id': None,
            'previous_quantity': 50.0,
            'market_value_in_base': None,
            'option_type': None,
            'book': None,
            'exchange': None,
            'error_message': None,
            'symbol': 'AAPL',
            'issuer': None,
            'trader': None,
            'extra_data': {'extra_field1': 'extra_value1',
            'extra_field2': 'extra_value2'},
            'maturity_date': '20250501',
            'series': None,
            'underlying': None,
            'quantity': 100.0,
            'percent_index': None,
            'fx_rate': None,
            'instrument_class': 'Equity',
            'market_value': None,
            'strike': None,
            'return_date': '20251231',
            'ticker': None,
            'isin': None,
            '__class_path__': 'everysk.sdk.entities.portfolio.security.Security',
        }

        self.assertIsInstance(result, dict)
        self.assertEqual(expected_data, result)

    def test_to_dict_without_internals_attributes(self):
        result = self.security.to_dict(add_class_path=False)
        expected_data = {
            'warranty': None,
            'currency': None,
            'operation': None,
            'market_price': None,
            'premium': None,
            'instrument_type': None,
            'status': 'OK',
            'cost_price': None,
            'error_type': None,
            'label': None,
            'unrealized_pl_in_base': None,
            'id': 'unique123',
            'instrument_subtype': None,
            'look_through_reference': None,
            'unrealized_pl': None,
            'multiplier': None,
            'issue_price': None,
            'comparable': 'Yes',
            'issue_date': '20200101',
            'display': 'Apple Inc.',
            'asset_class': None,
            'name': None,
            'rate': None,
            'issuer_type': None,
            'hash': None,
            'indexer': None,
            'settlement': '20200102',
            'accounting': None,
            'coupon': None,
            'asset_subclass': None,
            'trade_id': None,
            'previous_quantity': 50.0,
            'market_value_in_base': None,
            'option_type': None,
            'book': None,
            'exchange': None,
            'error_message': None,
            'symbol': 'AAPL',
            'issuer': None,
            'trader': None,
            'extra_data': {
                'extra_field1': 'extra_value1',
                'extra_field2': 'extra_value2'
            },
            'maturity_date': '20250501',
            'series': None,
            'underlying': None,
            'quantity': 100.0,
            'percent_index': None,
            'fx_rate': None,
            'instrument_class': 'Equity',
            'market_value': None,
            'strike': None,
            'return_date': '20251231',
            'ticker': None,
            'isin': None
        }

        self.assertIsInstance(result, dict)
        self.assertEqual(expected_data, result)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
__all__ = ['FileResponse', 'HTMLResponse', 'JSONResponse', 'PlainTextResponse', 'RedirectResponse', 'Response', 'StreamingResponse']

from typing import Any

from starlette.responses import Response, FileResponse, HTMLResponse, PlainTextResponse, RedirectResponse, StreamingResponse

from everysk.core.serialize import dumps


###############################################################################
#   JSONResponse Class Implementation
###############################################################################
class JSONResponse(Response):
    ## Private attributes
    _response_serializer: str = 'json'

    ## Public attributes
    media_type = 'application/json; charset=UTF-8'

    def render(self, content: Any) -> bytes:
        """
        Serialize the content to JSON format using the dumps function from the core.serialize module.

        Args:
            content (Any): The content to be serialized.
        """
        # with indent=None and separators=(",", ":"), the JSON will be minified
        return dumps(
            content,
            indent=None,
            separators=(",", ":"),
            protocol=self._response_serializer,
            use_undefined=True,
            return_type='bytes'
        )



# -*- coding: utf_8 -*-
###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.fields import StrField

from everysk.sdk.entities.base import BaseEntity
from everysk.sdk.entities.fields import EntityDescriptionField, EntityWorkspaceField

###############################################################################
#   Securities Field Implementation
###############################################################################

###############################################################################
#   Portfolio Implementation
###############################################################################
class Workspace(BaseEntity):

    name = EntityWorkspaceField()
    description = EntityDescriptionField(max_size=settings.WORKSPACE_DESCRIPTION_MAX_LENGTH)
    group = StrField(default=None, max_size=settings.WORKSPACE_GROUP_MAX_LENGTH, min_size=settings.WORKSPACE_GROUP_MIN_LENGTH, empty_is_none=True)

    @property
    def id(self) -> str:
        """
        Returns the id of the workspace. The id is a alias for the name.

        Returns:
            str: The id of the workspace.

        Example:
            >>> workspace = Workspace(name='custom')
            >>> workspace.id
            'custom'
        """
        return self.name

    @id.setter
    def id(self, value: str) -> None:
        """
        Sets the id of the workspace. The id is a alias for the name.

        Args:
            value (str): The id of the workspace.

        Example:
            >>> workspace = Workspace()
            >>> workspace.id = 'custom'
            >>> workspace.name
            'custom'
            >>> workspace.id
            'custom'
        """
        self.name = value

    @staticmethod
    def get_id_prefix():
        """
        Returns the id prefix for the workspace.

        Returns:
            str: The id prefix for the workspace.
        """
        return settings.WORKSPACE_ID_PREFIX

    def generate_id(self) -> str:
        """
        Generate a unique ID for an entity instance.

        Returns:
            str: The generated unique ID.

        Example:
            To generate a unique ID for an entity instance:
            >>> unique_id = MyEntity().generate_id()
        """
        raise NotImplementedError

    def to_dict(self, add_class_path = False, recursion = False) -> dict:
        """
        This method is used to convert the object to a dictionary.
        """
        dct: dict = super().to_dict(add_class_path, recursion)

        if not self.description:
            dct['description'] = ''

        return dct



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.datetime import DateTime, Date
from everysk.core.exceptions import FieldValueError
from everysk.core.unittests import TestCase

from everysk.sdk.entities.tags import Tags


###############################################################################
#   Tags TestCase Implementation
###############################################################################
class TagsTestCase(TestCase):

    def test_append_valid_tag(self):
        valid_tag = 'valid_tag'

        tags = Tags()
        tags.append(valid_tag)
        self.assertIn(valid_tag, tags)

    def test_append_invalid_tag(self):
        invalid_tag = {'a': 'b'}

        tags = Tags()
        with self.assertRaises(FieldValueError, msg="Tags: The string value does not match the required pattern"):
            tags.append(invalid_tag)

    def test_append_none_tag(self):
        tags = Tags()
        tags.append(None)
        self.assertListEqual(['none'], tags)

    def test_append_empty_tag(self):
        empty_tag = ''
        tags = Tags()
        tags.append(empty_tag)
        self.assertIn('none', tags)

    def test_append_float_tag(self):
        float_tag = 1.0
        tags = Tags()
        tags.append(float_tag)
        self.assertIn('1_0', tags)

    def test_append_long_tag(self):
        long_tag = 'a' * (settings.ENTITY_MAX_TAG_LENGTH + 1)
        tags = Tags()
        with self.assertRaises(FieldValueError, msg=f"Tags: '{long_tag}' size it's not between {tags.min_size} and {settings.ENTITY_MAX_TAG_LENGTH}"):
            tags.append(long_tag)

    def test_tag_sanitizes(self):
        tags = Tags(123)
        self.assertListEqual(['123'], tags)

        tags = Tags(123.321)
        self.assertListEqual(['123_321'], tags)

        tags = Tags(['tag'])
        self.assertListEqual(['tag'], tags)

        tags = Tags(['tag#123'])
        self.assertListEqual(['tag_123'], tags)

        tags = Tags(['Tag#123'])
        self.assertListEqual(['tag_123'], tags)

        tags = Tags(['T123!TT23'])
        self.assertListEqual(['t123_tt23'], tags)

        tags = Tags([123.321])
        self.assertListEqual(['123_321'], tags)

        #add a duplicate tag
        tags = Tags(['tag1', 'tag2', 'tag1'])
        self.assertListEqual(['tag1', 'tag2'], tags)

        tags = Tags('tag1')
        self.assertListEqual(['t', 'a', 'g', '1'], tags)

        tags = Tags(['tag1', 'tag2'])
        self.assertListEqual(['tag1', 'tag2'], tags)

        tags = Tags([['tag1', 'tag2'], ['tag1', 'tag3']])
        self.assertListEqual(['tag1', 'tag2', 'tag3'], tags)

        tags = Tags([['tag1', 'tag2'], ['tag1', 'tag3'], []])
        self.assertListEqual(['tag1', 'tag2', 'tag3'], tags)

        tags = Tags([[None, 'tag2'], ['tag1', Undefined, 'tag3'], ['', 'tag4']])
        self.assertListEqual(['none', 'tag2', 'tag1', 'tag3', 'tag4'], tags)

        # empty tag
        tags = Tags()
        self.assertListEqual([], tags)

        self.assertRaisesRegex(
            FieldValueError,
            "Unsupported format in Tags.",
            Tags,
            ['tag1', 'tag2', {'a': 'b'}]
        )

        self.assertRaisesRegex(
            FieldValueError,
            "Unsupported format in Tags.",
            Tags,
            [['tag1', 'tag2'], ['tag1', 'tag2', {'a': 'b'}]]
        )

    def test_extend_tags(self):
        tags = Tags(['tag1', 'tag2'])
        tags.extend(['tag3', 'tag4'])
        self.assertListEqual(['tag1', 'tag2', 'tag3', 'tag4'], tags)

        tags = Tags(['Tag#123', 'tag2'])
        tags.extend(['tag3', 'Tag!321'])
        self.assertListEqual(['tag_123', 'tag2', 'tag3', 'tag_321'], tags)

        #add a duplicate tag
        tags = Tags(['tag1', 'tag2'])
        tags.extend(['tag3', 'tag2'])
        self.assertListEqual(['tag1', 'tag2', 'tag3'], tags)

        tags = Tags(['tag1', 'tag2'])
        self.assertRaisesRegex(
            FieldValueError,
            "Unsupported format in Tags: <class 'dict'>",
            tags.extend,
            [{'a': 'b'}]
        )

    def test_insert(self):
        tags = Tags(['tag1', 'tag2'])
        tags.insert(0, 'tag3')
        self.assertListEqual(['tag3', 'tag1', 'tag2'], tags)

        tags = Tags(['tag1', 'tag2'])
        tags.insert(1, 'tag3')
        self.assertListEqual(['tag1', 'tag3', 'tag2'], tags)

        tags = Tags(['tag1', 'tag2'])
        self.assertRaisesRegex(
            FieldValueError,
            "Unsupported format in Tags: <class 'list'>",
            tags.insert,
            2,
            ['tag3', 'tag4']
        )

    def test__setitem__(self):
        tags = Tags(['tag1', 'tag2'])
        tags[0] = 'tag3'
        self.assertListEqual(['tag3', 'tag2'], tags)

        tags = Tags(['tag1', 'tag2'])
        tags[1] = 'tag3'
        self.assertListEqual(['tag1', 'tag3'], tags)

        tags = Tags(['tag1', 'tag2'])
        self.assertRaisesRegex(
            FieldValueError,
            "Unsupported format in Tags: <class 'list'>",
            tags.__setitem__,
            2,
            ['tag3', 'tag4']
        )

        class Fake:
            pass

        tags = Tags(['tag1', 'tag2'])
        self.assertRaisesRegex(
            FieldValueError,
            "Unsupported format in Tags: <class 'everysk.sdk.entities._tests.tags.TagsTestCase.test__setitem__.<locals>.Fake'>",
            tags.__setitem__,
            2,
            Fake()
        )

    def test_unify_tags(self):
        tags = Tags.unify([['tag1', 'tag2'], ['tag3', 'tag4']], ['tag5', 'tag6'])
        self.assertListEqual(['tag1', 'tag2', 'tag3', 'tag4'], tags)

        tags = Tags.unify([['tag1', 'tag2'], ['tag3', 'tag4']], [])
        self.assertListEqual(['tag1', 'tag2', 'tag3', 'tag4'], tags)

        tags = Tags.unify([['tag1', 'tag2'], ['tag3', 'tag4']], None)
        self.assertListEqual(['tag1', 'tag2', 'tag3', 'tag4'], tags)

        tags = Tags.unify([None, ['tag3', 'tag4']], ['tag1', 'tag2'])
        self.assertListEqual(['tag1', 'tag2', 'tag3', 'tag4'], tags)

        tags = Tags.unify([['tag1', 'tag2'], ['tag3', 'tag4']], ['tag1', 'tag2', 'tag3', 'tag4'])
        self.assertListEqual(['tag1', 'tag2', 'tag3', 'tag4'], tags)

        tags = Tags.unify([None], ['tag1', 'tag2'])
        self.assertListEqual(['tag1', 'tag2'], tags)

        tags = Tags.unify(None, ['tag1', 'tag2'])
        self.assertListEqual([], tags)

        tags = Tags.unify([], ['tag1', 'tag2'])
        self.assertListEqual([], tags)

        tags = Tags.unify(['tag1'], ['tag1', 'tag2'])
        self.assertListEqual(['tag1'], tags)

        tags = Tags.unify([['tag1']], ['tag1', 'tag2'])
        self.assertListEqual(['tag1'], tags)

        tags = Tags.unify([None], ['tag1', 'tag2'])
        self.assertListEqual(['tag1', 'tag2'], tags)

        tags = Tags.unify([[]], ['tag1', 'tag2'])
        self.assertListEqual(['tag1', 'tag2'], tags)

        tags = Tags.unify([DateTime(2023, 9, 9, 9, 9, 9, 999)], [None, 'tag1', 'tag2'])
        self.assertListEqual(['20230909'], tags)

        tags = Tags.unify([Date(2023, 9, 9)], [None, 'tag1', 'tag2'])
        self.assertListEqual(['20230909'], tags)

        tags = Tags.unify([True, False], [None, 'tag1', 'tag2'])
        self.assertListEqual(['true', 'false'], tags)

        tags = Tags.unify([[None]], [None, 'tag1', 'tag2'])
        self.assertListEqual(['none', 'tag1', 'tag2'], tags)

        tags = Tags.unify([[None]], ['tag1', 'tag2'])
        self.assertListEqual(['tag1', 'tag2'], tags)

        tags = Tags.unify(['tag3', ['tag3'], None, [], [None]], ['tag1', 'tag2'])
        self.assertListEqual(['tag3', 'tag1', 'tag2'], tags)

        tags = Tags.unify(['', [''], None], ['tag1', 'tag2'])
        self.assertListEqual(['tag1', 'tag2'], tags)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.datetime import DateTime
from everysk.core.http import HttpSDKPOSTConnection, httpx
from everysk.core.unittests import TestCase, mock

from everysk.sdk.entities.worker_execution.base import WorkerExecution, ParallelInfo, Result, InputParams

###############################################################################
#   WorkerExecution TestCase Implementation
###############################################################################
class WorkerExecutionTestCase(TestCase):

    def setUp(self):
        self.worker_execution: WorkerExecution = WorkerExecution(
            id='wkex_df0fcbd647552fb6dcd6046be',
            result=Result(data={'output1': 'output1', 'output2': 'output2'}),
            storage=False,
            parallel_info=ParallelInfo(),
            status='FAILED',
            execution_type='MANUAL',
            start_time=DateTime(2023, 9, 9, 9, 9, 9, 9),
            end_time=DateTime(2023, 9, 9, 9, 9, 9, 9),
            workflow_execution_id='wfex_1234567891234567891234567',
            workflow_id='wrkf_1234567891234567891234567',
            workflow_name='My Workflow',
            worker_id='wrkr_1234567891234567891234567',
            worker_name='My Worker',
            worker_type='BASIC',
            created_on=DateTime(2023, 9, 9, 9, 9, 9, 9),
            updated_on=DateTime(2023, 9, 9, 9, 9, 9, 9),
            input_params=InputParams(
                worker_id='wrkr_1234567891234567891234567',
                workflow_id='wrkf_1234567891234567891234567',
                worker_execution_id='wkex_df0fcbd647552fb6dcd6046be',
                workflow_execution_id='wfex_1234567891234567891234567',
                workspace='main',
                worker_type='BASIC',
                script_inputs={'input1': 'input1', 'input2': 'input2'},
                inputs_info={'input1': 'input1', 'input2': 'input2'},
                parallel_info=ParallelInfo(),
            )
        )

        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()

    def test_get_id_prefix(self):
        self.assertEqual(WorkerExecution.get_id_prefix(), settings.WORKER_EXECUTION_ID_PREFIX)

    def test_generate_id(self):
        self.assertRaisesRegex(
            NotImplementedError,
            '',
            WorkerExecution().generate_id
        )

    def test_query_load_with_id(self):
        expected_content = compress({'class_name': 'WorkerExecution', 'method_name': 'retrieve', 'self_obj': None, 'params': {'entity_id': 'wkex_df0fcbd647552fb6dcd6046be', 'projection': None}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            WorkerExecution(id='wkex_df0fcbd647552fb6dcd6046be', workflow_execution_id='wfex_1234567891234567891234567').load()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            content=expected_content
        )

    def test_query_load(self):
        worker_execution = WorkerExecution()
        expected_content = compress({'class_name': 'Query', 'method_name': 'load', 'self_obj': worker_execution.to_query().to_dict(add_class_path=True), 'params': {'offset': None}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            worker_execution.load()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            content=expected_content
        )

    def test_to_dict(self):
        result = self.worker_execution.to_dict()
        expected_content = {
            'input_params': {
                'inputs_info': {'input1': 'input1', 'input2': 'input2'},
                'parallel_info': {'length': -1, 'index': -1},
                'script_inputs': {'input1': 'input1', 'input2': 'input2'},
                'worker_execution_id': 'wkex_df0fcbd647552fb6dcd6046be',
                'worker_id': 'wrkr_1234567891234567891234567',
                'worker_type': 'BASIC',
                'workflow_execution_id': 'wfex_1234567891234567891234567',
                'workflow_id': 'wrkf_1234567891234567891234567',
                'workspace': 'main',
            },
            'result': {
                'output1': 'output1',
                'output2': 'output2'
            },
            'cpu_time': 0.0,
            'created': 1694250549,
            'duration': 0.0,
            'id': 'wkex_df0fcbd647552fb6dcd6046be',
            'parallel_info': {'length': -1, 'index': -1},
            'process_cpu_time': 0.0,
            'started': 1694250549,
            'status': 'FAILED',
            'trigger': 'MANUAL',
            'updated': 1694250549,
            'version': 'v1',
            'worker_id': 'wrkr_1234567891234567891234567',
            'worker_name': 'My Worker',
            'worker_type': 'BASIC',
            'workflow_execution_id': 'wfex_1234567891234567891234567',
            'workflow_id': 'wrkf_1234567891234567891234567',
            'workflow_name': 'My Workflow',
        }

        self.assertIsInstance(result, dict)
        self.assertEqual(result, expected_content)

        self.worker_execution.duration = None
        result = self.worker_execution.to_dict()

        self.assertIsInstance(result, dict)
        self.assertEqual(result, expected_content)

    def test_get_input_params(self):
        self.assertRaisesRegex(
            ValueError,
            'Invalid Entity ID: invalid_id',
            WorkerExecution.get_input_params,
            'invalid_id'
        )

        with mock.patch('everysk.sdk.entities.worker_execution.base.WorkerExecution.retrieve', return_value=None) as mock_retrieve:
            self.assertRaisesRegex(
                ValueError,
                'Entity not found. Entity ID: wkex_df0fcbd647552fb6dcd6046be',
                WorkerExecution.get_input_params,
                'wkex_df0fcbd647552fb6dcd6046be'
            )

        mock_retrieve.assert_called_once_with('wkex_df0fcbd647552fb6dcd6046be')

        with mock.patch('everysk.sdk.entities.worker_execution.base.WorkerExecution.retrieve') as mock_retrieve:
            mock_retrieve.return_value = WorkerExecution(
                id='wkex_df0fcbd647552fb6dcd6046be',
                input_params=InputParams(),
                created_on=DateTime(2023, 9, 9, 9, 9, 9, 9),
                updated_on=DateTime(2023, 9, 9, 9, 9, 9, 9)
            )
            result = WorkerExecution.get_input_params('wkex_df0fcbd647552fb6dcd6046be')

        mock_retrieve.assert_called_once_with('wkex_df0fcbd647552fb6dcd6046be')
        self.assertEqual(
            result,
            InputParams(worker_execution_id='wkex_df0fcbd647552fb6dcd6046be')
        )

        with mock.patch('everysk.sdk.entities.worker_execution.base.WorkerExecution.retrieve') as mock_retrieve:
            mock_retrieve.return_value = WorkerExecution(
                id='wkex_df0fcbd647552fb6dcd6046be',
                input_params=InputParams(worker_execution_id='wkex_df0fcbd647552fb6dcd6046bc'),
                created_on=DateTime(2023, 9, 9, 9, 9, 9, 9),
                updated_on=DateTime(2023, 9, 9, 9, 9, 9, 9)
            )
            result = WorkerExecution.get_input_params('wkex_df0fcbd647552fb6dcd6046be')

        mock_retrieve.assert_called_once_with('wkex_df0fcbd647552fb6dcd6046be')
        self.assertEqual(
            result,
            InputParams(worker_execution_id='wkex_df0fcbd647552fb6dcd6046bc')
        )

    def test_to_query(self):
        result = WorkerExecution(workflow_execution_id='wfex_1234567891234567891234567').to_query()
        self.assertEqual(result, WorkerExecution.query.where('workflow_execution_id', 'wfex_1234567891234567891234567'))

    def test_init(self):
        self.assertEqual(
            WorkerExecution(
                result={'data': [1,2,3,4]},
                input_params={'workflow_execution_id': 'wfex_1234567891234567891234567'},
                parallel_info={'index': -1, 'length': -1},
                created_on=DateTime(2023, 9, 9, 9, 9, 9, 9),
                updated_on=DateTime(2023, 9, 9, 9, 9, 9, 9)
            ),
            WorkerExecution(
                result=Result(data= [1,2,3,4]),
                input_params=InputParams(workflow_execution_id='wfex_1234567891234567891234567'),
                parallel_info=ParallelInfo(index=-1, length=-1),
                created_on=DateTime(2023, 9, 9, 9, 9, 9, 9),
                updated_on=DateTime(2023, 9, 9, 9, 9, 9, 9)
            )
        )



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=unused-import

## Remember to prefix all import with EveryskLib to avoid clash with other tests
try:
    from everysk.sdk._tests.base import TestBaseSDK as EveryskLibTestBaseSDK
    from everysk.sdk._tests.init import SDKInitTestCase as EveryskLibSDKInitTestCase
    from everysk.sdk._tests.serialize import (
        SerializeDumpsSDKTestCase as EveryskLibSerializeDumpsSDKTestCase,
        SerializeLoadsSDKTestCase as EveryskLibSerializeLoadsSDKTestCase
    )
    from everysk.sdk._tests.worker_base import WorkerBaseTestCase as EveryskLibWorkerBaseTestCase

    from everysk.sdk.entities._tests.init import InitTestCase as EveryskLibInitTestCase


    from everysk.sdk.engines._tests.cryptography import TestCryptography as EveryskLibTestCryptography
    from everysk.sdk.engines._tests.init import EnginesInitTestCase as EveryskLibEnginesInitTestCase
    from everysk.sdk.engines._tests.compliance import ComplianceTestCase as EveryskLibComplianceTestCase
    from everysk.sdk.engines._tests.expression import ExpressionTestCase as EveryskLibExpressionTestCase
    from everysk.sdk.engines._tests.cache import CacheTestCase as EveryskLibCacheTestCase
    from everysk.sdk.engines._tests.market_data import MarketDataTestCase as EveryskLibMarketDataTestCase

    from everysk.sdk.entities._tests.base_list import TestEntityList as EveryskLibTestEntityList

    from everysk.sdk.entities._tests.script import ScriptTestCase as EveryskLibScriptTestCase
    from everysk.sdk.entities._tests.base import TestBaseEntity as EveryskLibTestBaseEntity
    from everysk.sdk.entities._tests.tags import TagsTestCase as EveryskLibTestTagsList
    from everysk.sdk.entities._tests.query import QueryTestCase as EveryskLibQueryTestCase

    from everysk.sdk.entities.portfolio._tests.security import TestSecurity as EveryskLibTestSecurity
    from everysk.sdk.entities.portfolio._tests.securities import TestSecurities as EveryskLibTestSecurities
    from everysk.sdk.entities.portfolio._tests.base import (
        TestSecuritiesField as EveryskLibTestTestSecuritiesField,
        TestPortfolio as EveryskLibTestPortfolio
    )

    from everysk.sdk.entities._tests.fields import (
        TestBaseCurrencyField as EveryskLibTestBaseCurrencyField,
        TestEntityTagsField as EveryskLibTestEntityTagsField,
        TestEntityNameField as EveryskLibTestEntityNameField,
        TestEntityDescriptionField as EveryskLibTestEntityDescriptionField,
        TestEntityLinkUIDField as EveryskLibTestEntityLinkUIDField,
        TestEntityWorkspaceField as EveryskLibTestEntityWorkspaceField,
        TestEntityDateTimeField as EveryskLibTestEntityDateTimeField
    )

    from everysk.sdk.entities.custom_index._tests.base import CustomIndexTestCase as EveryskLibCustomIndexTestCase
    from everysk.sdk.entities.datastore.tests.base import DatastoreTestCase as EveryskLibDatastoreTestCase
    from everysk.sdk.entities.private_security._tests.base import PrivateSecurityTestCase as EveryskLibPrivateSecurityTestCase
    from everysk.sdk.entities.file._tests.base import FileTestCase as EveryskLibFileTestCase
    from everysk.sdk.entities.report.tests.base import ReportTestCase as EveryskLibReportTestCase
    from everysk.sdk.entities.worker_execution.tests.base import WorkerExecutionTestCase as EveryskLibWorkerExecutionTestCase
    from everysk.sdk.entities.workflow_execution.tests.base import WorkflowExecutionTestCase as EveryskLibWorkflowExecutionTestCase
    from everysk.sdk.entities.workspace._tests.base import TestWorkspace as EveryskLibTestWorkspace

except ModuleNotFoundError as error:
    # This will prevent running these tests if requests is not installed
    if not error.args[0].startswith("No module named 'requests'"):
        raise error



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core.fields import IntField, StrField, BoolField, DictField


SIMPLE_UNIQUE_ID_LENGTH = IntField(default=8, readonly=True)

EVERYSK_API_SID = StrField()
EVERYSK_API_TOKEN = StrField()

EVERYSK_APP_URL_SCHEME = StrField(default='https')
EVERYSK_APP_URL_DOMAIN = StrField(default='app.everysk.com')
EVERYSK_APP_URL = StrField(default='{EVERYSK_APP_URL_SCHEME}://{EVERYSK_APP_URL_DOMAIN}')

EVERYSK_API_URL_SCHEME = StrField(default='https')
EVERYSK_API_URL_DOMAIN = StrField(default='api.everysk.com')
EVERYSK_API_URL = StrField(default='{EVERYSK_API_URL_SCHEME}://{EVERYSK_API_URL_DOMAIN}')
EVERYSK_SDK_URL = EVERYSK_API_URL

EVERYSK_API_VERSION = StrField(default='v2')
EVERYSK_API_VERIFY_SSL_CERTS = BoolField(default=True)

EVERYSK_SDK_VERSION = StrField(default='v1', readonly=True)
EVERYSK_SDK_ROUTE = StrField(default='sdk_function', readonly=True)

EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT = IntField(default=3600, readonly=True)

EVERYSK_SDK_MODULES_PATH = DictField(
    default={
        'WorkerBase': 'everysk.sdk.worker_base.WorkerBase'
    },
    readonly=True
)

EVERYSK_SDK_ENTITIES_MODULES_PATH = DictField(
    default={
        'CustomIndex': 'everysk.sdk.entities.custom_index.base.CustomIndex',
        'Datastore': 'everysk.sdk.entities.datastore.base.Datastore',
        'File': 'everysk.sdk.entities.file.base.File',
        'Portfolio': 'everysk.sdk.entities.portfolio.base.Portfolio',
        'PrivateSecurity': 'everysk.sdk.entities.private_security.base.PrivateSecurity',
        'Query': 'everysk.sdk.entities.query.Query',
        'Report': 'everysk.sdk.entities.report.base.Report',
        'Script': 'everysk.sdk.entities.script.Script',
        'Securities': 'everysk.sdk.entities.portfolio.securities.Securities',
        'Security': 'everysk.sdk.entities.portfolio.security.Security',
        'WorkerExecution': 'everysk.sdk.entities.worker_execution.base.WorkerExecution',
        'WorkflowExecution': 'everysk.sdk.entities.workflow_execution.base.WorkflowExecution',
        'Workspace': 'everysk.sdk.entities.workspace.base.Workspace'
    },
    readonly=True
)

EVERYSK_SDK_ENGINES_MODULES_PATH = DictField(
    default={
        'Expression': 'everysk.sdk.engines.expression.Expression',
        'Compliance': 'everysk.sdk.engines.compliance.Compliance',
        'UserCache': 'everysk.sdk.engines.cache.UserCache',
        'MarketData': 'everysk.sdk.engines.market_data.MarketData',
    },
    readonly=True
)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core.fields import StrField, ListField, IntField, RegexField


WORKER_EXECUTION_ID_PREFIX = StrField(default='wkex_', readonly=True)
WORKER_EXECUTION_ID_LENGTH = IntField(default=25, readonly=True)
WORKER_EXECUTION_ID_REGEX = RegexField(default=r'^wkex_[a-zA-Z0-9]{25}', readonly=True)

WORKER_ID_PREFIX = StrField(default='wrkr_', readonly=True)
WORKER_ID_LENGTH = IntField(default=25, readonly=True)
WORKER_ID_REGEX = RegexField(default=r'^wrkr_[a-zA-Z0-9]{25}', readonly=True)

WORKER_EXECUTION_UNFORKED_PARALLEL_INDEX = IntField(default=-1, readonly=True)
WORKER_EXECUTION_UNFORKED_PARALLEL_LENGTH = IntField(default=-1, readonly=True)

WORKER_EXECUTION_STATUS_COMPLETED = StrField(default='COMPLETED', readonly=True)
WORKER_EXECUTION_STATUS_FAILED = StrField(default='FAILED', readonly=True)
WORKER_EXECUTION_STATUS_PREPARING = StrField(default='PREPARING', readonly=True)
WORKER_EXECUTION_STATUS_RUNNING = StrField(default='RUNNING', readonly=True)
WORKER_EXECUTION_STATUS_LIST = ListField(default=['COMPLETED', 'FAILED', 'PREPARING', 'RUNNING'], readonly=True)

WORKER_EXECUTION_INTEGRATION_EVENT_TYPE = StrField(default='INTEGRATION_EVENT', readonly=True)
WORKER_EXECUTION_TIME_EVENT_TYPE = StrField(default='TIME_EVENT', readonly=True)
WORKER_EXECUTION_SCHEDULER_EVENT_TYPE = StrField(default='SCHEDULER_EVENT', readonly=True)
WORKER_EXECUTION_WORKER_FINISHED_EVENT_TYPE = StrField(default='WORKER_FINISHED_EVENT', readonly=True)
WORKER_EXECUTION_MANUAL_EVENT_TYPE = StrField(default='MANUAL', readonly=True)
WORKER_EXECUTION_API_EVENT_TYPE = StrField(default='API', readonly=True)
WORKER_EXECUTION_EVENT_TYPE_LIST = ListField(default=['INTEGRATION_EVENT', 'TIME_EVENT', 'SCHEDULER_EVENT', 'WORKER_FINISHED_EVENT', 'MANUAL', 'API'], readonly=True)

WORKER_EXECUTION_TYPE_TIME_DRIVEN = StrField(default='TIME_DRIVEN', readonly=True)
WORKER_EXECUTION_TYPE_EVENT_DRIVEN = StrField(default='EVENT_DRIVEN', readonly=True)
WORKER_EXECUTION_TYPE_MANUAL = StrField(default='MANUAL', readonly=True)
WORKER_EXECUTION_TYPE_API = StrField(default='API', readonly=True)
WORKER_EXECUTION_TYPE_LIST = ListField(default=['TIME_DRIVEN', 'EVENT_DRIVEN', 'SCHEDULER_EVENT', 'MANUAL', 'API'], readonly=True)

RESULT_STATUS_OK = StrField(default='OK', readonly=True)
RESULT_STATUS_ERROR = StrField(default='ERROR', readonly=True)

WORKER_TEMPLATE_TYPES = ListField(
    default=['STARTER', 'BASIC', 'FORKER', 'BARRIER', 'CONDITIONAL', 'TRY', 'CATCH', 'ENDER', 'Retrieve'], # Retrieve is a legacy type.
    readonly=True
)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.http import HttpSDKPOSTConnection, httpx
from everysk.core.unittests import TestCase, mock
from everysk.core.compress import compress

from everysk.sdk.entities.workspace.base import Workspace
###############################################################################
#   Workspace TestCase Implementation
###############################################################################
class TestWorkspace(TestCase):

    def setUp(self):
        self.sample_data = {
            'name': 'main',
            'description': 'Description',
            'group': 'Test'
        }
        self.workspace = Workspace(**self.sample_data)
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()

    def test_static_methods(self):
        self.assertEqual(Workspace.get_id_prefix(), settings.WORKSPACE_ID_PREFIX)

    def test_query_load_with_id(self):
        workspace = Workspace(id='main', group='Test')
        expected_content = compress({'class_name': 'Workspace', 'method_name': 'retrieve', 'self_obj': None, 'params': {'entity_id': 'main', 'projection': None}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            workspace.load()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            content=expected_content
        )

    def test_generate_id(self):
        self.assertRaisesRegex(
            NotImplementedError,
            '',
            Workspace().generate_id
        )

    def test_to_dict(self):
        dct = self.workspace.to_dict()
        self.assertEqual(dct['description'], 'Description')
        self.assertEqual(dct['group'], 'Test')
        self.assertEqual(dct['name'], 'main')

        self.workspace.description = None
        dct = self.workspace.to_dict()
        self.assertEqual(dct['description'], '')
        self.assertEqual(dct['group'], 'Test')
        self.assertEqual(dct['name'], 'main')

    def test_workspace_id(self):
        self.assertEqual(self.workspace.id, 'main')
        self.workspace.id = 'main2'
        self.assertEqual(self.workspace.name, 'main2')
        self.assertEqual(self.workspace.id, 'main2')
        self.workspace.name = 'main3'
        self.assertEqual(self.workspace.id, 'main3')
        self.assertEqual(self.workspace.name, 'main3')
        self.workspace.id = None
        self.assertEqual(self.workspace.id, None)
        self.assertEqual(self.workspace.name, None)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core.fields import StrField, IntField, FloatField, RegexField, ListField


PORTFOLIO_ID_REGEX = RegexField(default=r'^port_[a-zA-Z0-9]', readonly=True)
PORTFOLIO_ID_PREFIX = StrField(default='port_', readonly=True)
PORTFOLIO_ID_MAX_SIZE = IntField(default=30, readonly=True) # len(PORTFOLIO_ID_PREFIX) + ENTITY_ID_LENGTH

PORTFOLIO_MAX_SIZE = IntField(default=30000, readonly=True)
PORTFOLIO_LEVEL_MAX_LENGTH = IntField(default=32, readonly=True)

SECURITY_STATUS_OK = StrField(default='OK', readonly=True)
SECURITY_STATUS_ERROR = StrField(default='ERROR', readonly=True)
SECURITY_STATUS_DELISTED = StrField(default='DELISTED', readonly=True)
SECURITY_ID_LENGTH = IntField(default=6, readonly=True)
SECURITY_ID_PREFIX = StrField(default='sec_', readonly=True)
SYMBOL_ID_MAX_LEN = IntField(default=100, readonly=True)

PRICE_MIN_VALUE = FloatField(default=-9999999999999.99, readonly=True)
PRICE_MAX_VALUE = FloatField(default=+9999999999999.99, readonly=True)

FX_RATE_MIN_VALUE = FloatField(default=0.0, readonly=True)
FX_RATE_MAX_VALUE = FloatField(default=+9999999999999.99, readonly=True)

MULTIPLIER_MIN_VALUE = FloatField(default=0.0, readonly=True)
MULTIPLIER_MAX_VALUE = FloatField(default=+9999999999999.99, readonly=True)

QUANTITY_MIN_VALUE = FloatField(default=-9999999999999.99, readonly=True)
QUANTITY_MAX_VALUE = FloatField(default=+9999999999999.99, readonly=True)

SYMBOL_MAX_LENGTH = IntField(default=100, readonly=True)
LABEL_MAX_LENGTH = IntField(default=100, readonly=True)

PORTFOLIO_SECURITY_ERROR_TYPE_MAX_LEN = IntField(default=100, readonly=True)
PORTFOLIO_SECURITY_ERROR_MESSAGE_MAX_LEN = IntField(default=500, readonly=True)
PORTFOLIO_SECURITY_TYPE_MAX_LEN = IntField(default=100, readonly=True)

PORTFOLIO_PROPERTIES_ORDER = ListField(
    default=[
        'status',
        'id',
        'symbol',
        'quantity',
        'instrument_class',
        'ticker',
        'label',
        'name',
        'isin',
        'exchange',
        'currency',
        'fx_rate',
        'market_price',
        'market_value',
        'instrument_type',
        'instrument_subtype',
        'asset_class',
        'asset_subclass',
        'error_type',
        'error_message',
        'maturity_date',
        'indexer',
        'percent_index',
        'rate',
        'coupon',
        'multiplier',
        'underlying',
        'series',
        'option_type',
        'strike',
        'issue_price',
        'issue_date',
        'issuer',
        'issuer_type',
        'cost_price',
        'unrealized_pl',
        'unrealized_pl_in_base',
        'book',
        'trader',
        'trade_id',
        'operation',
        'accounting',
        'warranty',
        'return_date',
        'settlement',
        'look_through_reference',
        'extra_data',
        'hash'
    ],
    readonly=True
)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.core.fields import IntField, RegexField, StrField

###############################################################################
#   Settings Implementation
###############################################################################
ENTITY_WORKSPACE_MAX_LENGTH = IntField(default=50, readonly=True)
ENTITY_WORKSPACE_MIN_LENGTH = IntField(default=1, readonly=True)
ENTITY_WORKSPACE_REGEX = RegexField(default=r'^[a-zA-Z0-9_]*$', readonly=True)

WORKSPACE_GROUP_MAX_LENGTH = IntField(default=50, readonly=True)
WORKSPACE_GROUP_MIN_LENGTH = IntField(default=1, readonly=True)

WORKSPACE_ID_PREFIX = StrField(default='wksp_', readonly=True)
WORKSPACE_DESCRIPTION_MAX_LENGTH = IntField(default=100, readonly=True)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core.exceptions import SigningError
from everysk.core.serialize import dumps, loads
from everysk.core.unittests import TestCase, mock


class SerializePickleDumpsTestCase(TestCase):

    @mock.patch('os.environ', {'EVERYSK_SIGNING_KEY': 'test'})
    def test_protocol_pickle(self):
        self.assertEqual(
            dumps(1, protocol='pickle'),
            b'68cec7e4c81fe8571641ae3709d4f546841ec9b1:\x80\x04K\x01.'
        )

    @mock.patch('os.environ', {'EVERYSK_SIGNING_KEY': None})
    def test_protocol_pickle_no_key(self):
        self.assertEqual(
            dumps(1, protocol='pickle'),
            b'\x80\x04K\x01.'
        )


class SerializePickleLoadsTestCase(TestCase):

    @mock.patch('os.environ', {'EVERYSK_SIGNING_KEY': 'test'})
    def test_protocol_pickle(self):
        self.assertEqual(
            loads(b'68cec7e4c81fe8571641ae3709d4f546841ec9b1:\x80\x04K\x01.', protocol='pickle'),
            1
        )

    @mock.patch('os.environ', {'EVERYSK_SIGNING_KEY': 'test_invalid'})
    def test_protocol_pickle_invalid(self):
        with self.assertRaisesRegex(SigningError, 'Error trying to unsign data.'):
            loads(b'1dd5f31b5032a28d51025f46121b03ebe20f258a:\x80\x04\x95\x05\x00\x00\x00\x00\x00\x00\x00\x8c\x01k\x94.', protocol='pickle')

    @mock.patch('os.environ', {'EVERYSK_SIGNING_KEY': ''})
    def test_protocol_pickle_no_key(self):
        self.assertEqual(loads(b'\x80\x04K\x01.', protocol='pickle'), 1)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import requests

from everysk.core.serialize import dumps
from everysk.core.datetime.date_settings import DEFAULT_DATE_FORMAT, DEFAULT_DATE_TIME_FORMAT

from requests.packages.urllib3.exceptions import InsecureRequestWarning # pylint: disable=import-error


requests.packages.urllib3.disable_warnings(InsecureRequestWarning) # pylint: disable=no-member


###############################################################################
#   HTTPClient Implementation
###############################################################################
class HTTPClient(object):
    def __init__(self, timeout=3600, verify_ssl_certs=True, allow_redirects=False) -> None:
        self.timeout = timeout
        self.allow_redirects = allow_redirects
        self.verify_ssl_certs = verify_ssl_certs

    def request(self, method, url, headers, params=None, payload=None):
        raise NotImplementedError(
            'HTTPClient subclasses must implement `request`.'
        )


###############################################################################
#   Requests Client Implementation
###############################################################################
class RequestsClient(HTTPClient):
    def request(self, method, url, headers, params=None, payload=None):

        _args = {
            'method': method,
            'url': url,
            'headers': headers,
            'params': params,
            'timeout': self.timeout,
            'verify': self.verify_ssl_certs,
            'allow_redirects': self.allow_redirects
        }

        if payload:
            _args['data'] = dumps(payload, date_format=DEFAULT_DATE_FORMAT, datetime_format=DEFAULT_DATE_TIME_FORMAT, use_undefined=False, add_class_path=False)

        response = requests.request(**_args)

        return (response.status_code, response.content)

def new_default_http_client(*args, **kwargs):
    return RequestsClient(*args, **kwargs)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.core.fields import IntField, StrField, SetField

###############################################################################
#   Settings Implementation
###############################################################################
ENTITY_NAME_MAX_LENGTH = IntField(default=200, readonly=True)
ENTITY_NAME_MIN_LENGTH = IntField(default=1, readonly=True)

ENTITY_DESCRIPTION_MAX_LEN = IntField(default=500, readonly=True)
ENTITY_DESCRIPTION_MIN_LEN = IntField(default=0, readonly=True)

ENTITY_MAX_TAG_LENGTH = IntField(default=252, readonly=True)
ENTITY_MIN_TAG_LENGTH = IntField(default=1, readonly=True)
ENTITY_MAX_TAG_SIZE = IntField(default=32, readonly=True)
ENTITY_MIN_TAG_SIZE = IntField(default=0, readonly=True)

ENTITY_DEFAULT_VERSION = StrField(default='v1', readonly=True)
ENTITY_ID_LENGTH = IntField(default=25, readonly=True)

ENTITY_MAX_TAG_SIZE = IntField(default=32, readonly=True)
ENTITY_MIN_TAG_SIZE = IntField(default=0, readonly=True)

ENTITY_LINK_UID_MAX_LENGTH = IntField(default=128, readonly=True)
ENTITY_LINK_UID_MIN_LENGTH = IntField(default=1, readonly=True)

DEFAULT_QUERY_OFFSET = IntField(default=0, readonly=True)
DEFAULT_QUERY_PAGE_SIZE = IntField(default=20, readonly=True)
DEFAULT_QUERY_LIMIT = IntField(default=None)

ENTITY_RETRIEVE_BATCH_SIZE = IntField(default=1000, readonly=True)

QUERY_OPERATORS = SetField(default={'<=', '>=', '<', '>', '=', '!=', 'IN', 'NOT_IN'}, readonly=True)

ENTITY_BASE_CURRENCY_DEFAULT_LIST = SetField(default={'USD', 'EUR', 'GBP', 'JPY', 'CNY', 'AUD', 'CAD', 'CHF', 'CNH', 'HKD', 'NZD', 'BRL'}, readonly=True)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import re
from typing import Any
from re import compile as compile_re
from importlib import import_module
from unicodedata import normalize
from six import string_types, ensure_str, text_type


RE_ISIN = compile_re(r'^[a-zA-Z]{2}\s*[0-9a-zA-Z]{9}[0-9](?![0-9a-zA-Z-])$')


###############################################################################
#   Public Functions Implementation
###############################################################################
def import_from_string(dotted_path: str) -> type:
    """
    Import and return a Python class or module based on a dotted path.

    Args:
        dotted_path (str): The dotted path to the Python class or module.

    Returns:
        type: The imported Python class or module.

    Raises:
        ImportError: If the provided dotted path is invalid or if the class/module cannot be imported.

    Example:
        >>> import_from_string("module_name.ClassName")
        <class 'module_name.ClassName'>
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as error:
        raise ImportError(f"{dotted_path} doesn't look like a module path.") from error

    module = import_module(module_path)
    try:
        return getattr(module, class_name)
    except AttributeError as error:
        raise ImportError(f'Module "{module_path}" does not define a "{class_name}" class.') from error

def is_string_object(obj: Any) -> bool:
    """
    Check if the provided object is a string object.

    Args:
        obj (Any): The object to check.

    Returns:
        bool: True if the object is a string, False otherwise.

    Example:
        >>> from everysk.core.string import is_string_object
        >>> is_string_object('Hello, World!')
        >>> True

        >>> is_string_object(42)
        >>> False
    """
    return isinstance(obj, string_types)

def to_string(value: str) -> str:
    """
    Convert a value to a string representation.

    Args:
        value (str): The value to convert to a string.

    Returns:
        str: The string representation of the value.

    Example:
        Convert a value to a string:
        >>> to_string(42)
        '42'
    """
    if is_string_object(value):
        return ensure_str(value)
    return text_type(value)

def normalize_string(value: str):
    """
    Normalize a string by applying Unicode normalization (NFKD).

    Args:
        value (str): The string to normalize.

    Returns:
        str: The normalized string.

    Example:
        >>> from everysk.core.string import normalize_string
        >>> result = normalize_string("Café")
        >>> print(result)
        >>> "Cafe"

    """
    value = normalize('NFKD', value)
    return value

def normalize_string_to_search(value):
    """
    Normalize a string for search purposes by lowercasing and applying Unicode normalization (NFKD).

    Args:
        value: The string to normalize.

    Returns:
        str: The normalized and lowercase string suitable for searching.
    """
    value = value.lower().strip()
    value = to_string(value)
    return normalize_string(value)

def is_isin_string_regex(isin: str) -> bool:
    """
    Check if a provided string matches the ISIN (International Securities Identification Number) format using regex.

    Args:
        isin (str): The string to check.

    Returns:
        bool: True if the string matches the ISIN format, False otherwise.

    Example:
        >>> from everysk.core.string import is_isin_string_regex
        >>> result = is_isin_string_regex("US0378331005")
        >>> print(result)
        >>> True
    """
    if not isinstance(isin, string_types):
        return False
    return True if RE_ISIN.match(isin) else False

def pluralize(string: str) -> str:
    """
    Transform the string into its pluralized version.

    Args:
        string (str): The string that we want to pluralize.

    Example:
        >>> from.everysk.core.string import pluralize
        >>> pluralize('study')
        'studies'

        >>> pluralize('class')
        'classes'

        >>> pluralize('tax')
        'taxes'

        >>> pluralize('system')
        'systems'

    Returns:
        string (str): The pluralized version of the string
    """
    if string[-1] in ('s', 'x'):
        string += 'es'
    elif string[-1] in ('y'):
        string = string[:-1] + 'ies'
    else:
        string += 's'
    return string

def snake_case(string: str) -> str:
    """
    Convert the current string into the snake case format.

    Args:
        string (str): The string for conversion

    Example:
        >>> from everysk.core.string import snake_case
        >>> snake_case('CustomIndex')
        'custom_index'

        >>> snake_case('Security')
        'security'

        >>> snake_case('ClassOrObjectByTest')
        'class_or_object_by_test'

    Returns:
        string (str): The string converted to to snake case format
    """
    _snake_case = re.compile(r'(?<!^)(?=[A-Z])')
    return _snake_case.sub('_', string).lower()
